import copy

from libfaketime import fake_time
from slimta.smtp.reply import Reply
from django.conf import settings
from django.test import override_settings

from munch_mailsend.models import Mail
from munch_mailsend.models import MailStatus
from munch_mailsend import policies
from munch_mailsend.models import Worker
from munch_mailsend.policies.mx import greylist

from . import MailSendTestCase


class GreyListPolicyTestCase(MailSendTestCase):
    def test_greylisting_kk(self):
        worker_01 = Worker.objects.create(
            name='worker_01', ip='10.0.0.1',
            policies_settings={'greylist': {}})
        Worker.objects.create(
            name='worker_02', ip='10.0.0.2',
            policies_settings={'greylist': {}})

        mail_01 = Mail.objects.create(
            identifier='0001',
            headers={
                'To': 'you@example.com',
                'X-MAILSEND-Message-Id': '0001'})
        with fake_time('2015-12-10 12:00:00'):
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_01, source_ip=worker_01.ip)
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_01, status=MailStatus.SENDING,
                source_ip=worker_01.ip)
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_01, status=MailStatus.DELAYED,
                source_ip=worker_01.ip)

        with fake_time('2015-12-10 12:05:00'):
            MAILSEND_SETTINGS = copy.deepcopy(settings.MAILSEND)
            MAILSEND_SETTINGS['WORKER_POLICIES'] = [
                'munch_mailsend.policies.mx.greylist.Policy']
            with override_settings(MAILSEND=MAILSEND_SETTINGS):
                best_worker, next_available, score, *_ = Worker.objects.\
                    find_worker(
                        mail_01.identifier, mail_01.headers, MailStatus,
                        reply=Reply(code='420', message='4.2.0 GreyListedd'))
            self.assertEqual(best_worker.name, worker_01.name)
            self.assertEqual(best_worker.ip, worker_01.ip)
            self.assertEqual(score, 1.0)

    def test_greylisting_not_delayed(self):
        worker_01 = Worker.objects.create(
            name='worker_01', ip='10.0.0.1',
            policies_settings={'greylist': {}})
        Worker.objects.create(
            name='worker_02', ip='10.0.0.2',
            policies_settings={'greylist': {}})

        mail_metadata_01 = Mail.objects.create(identifier='0001')
        with fake_time('2015-12-10 12:00:00'):
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_01, source_ip=worker_01.ip)
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_01,
                status=MailStatus.SENDING,
                source_ip=worker_01.ip)
            MailStatus.objects.create(
                destination_domain='example.com', mail=mail_metadata_01,
                status=MailStatus.DELIVERED, source_ip=worker_01.ip)

        with fake_time('2015-12-10 12:05:00'):
            workers = policies.mx.First().apply({'To': 'test+01@example.com'})
            worker_ranking = greylist.Policy(
                mail_metadata_01.identifier, {'To': 'test+01@example.com'},
                MailStatus, reply_code='4.2.0',
                reply_message='GreylisTeDd').apply(workers)

            self.assertEqual(worker_ranking[0]['score'], 0.0)
            self.assertEqual(worker_ranking[1]['score'], 0.0)
            best_worker, next_available, *_ = policies.mx.Last().apply(
                worker_ranking)



from django.conf import settings

settings.MAILSEND['X_POOL_HEADER'] = 'X-MAILSEND-Pool'
settings.MAILSEND['WORKER_POLICIES_SETTINGS'] = {
    'rate_limit': {'domains': [(r'.*', 2)], 'max_queued': 60 * 15},
    'warm_up': {
        'prioritize': 'equal',
        'domain_warm_up': {
            'matrice': [50, 100, 300, 500, 1000],
            'goal': 500,
            'max_tolerance': 10,
            'step_tolerance': 10,
            'domains_list': []},
        'ip_warm_up': {
            'matrice': [50, 100, 300, 500, 1000],
            'goal': 500,
            'max_tolerance': 10,
            'step_tolerance': 10,
            'enabled': False
        }
    }
}



import logging

from django.conf import settings
from django.core.mail.backends.base import BaseEmailBackend
from django.utils.module_loading import import_string

from munch.core.mail.utils import extract_domain
from munch.core.mail.utils import mk_base64_uuid
from munch.core.mail.models import AbstractMailStatus

from .models import Mail
from .models import RawMail
from .tasks import route_envelope
from .utils import message_to_envelope

log = logging.getLogger(__name__)


class Backend(BaseEmailBackend):
    def __init__(
            self,
            build_envelope_task_path='mailsend.utils.tasks.get_envelope',
            mailstatus_class_path='mailsend.models.MailStatus',
            record_status_task_path='mailsend.utils.tasks.record_status',
            *args, **kwargs):
        self.mailstatus_class_path = mailstatus_class_path
        self.record_status_task_path = record_status_task_path
        self.build_envelope_task_path = build_envelope_task_path

        self.get_envelope = import_string(self.build_envelope_task_path)
        self.mailstatus_class = import_string(self.mailstatus_class_path)
        self.record_status_task = import_string(self.record_status_task_path)

        self.sandbox = settings.MAILSEND['SANDBOX']

    def _route(self, identifier, headers, attempts, priority=50):
        return route_envelope.apply_async(
            (identifier, headers, attempts), {
                'mailstatus_class_path': self.mailstatus_class_path,
                'record_status_task_path': self.record_status_task_path,
                'build_envelope_task_path': self.build_envelope_task_path},
            priority=priority).id

    def _handle_sandbox(self, identifier, recipient):
        log.info('Ignoring {} envelope because SANDBOX is enabled'.format(
            identifier))
        mailstatus = self.mailstatus_class(
            status=AbstractMailStatus.SENDING,
            destination_domain=extract_domain(recipient))
        self.record_status_task(
            mailstatus, identifier,
            settings.MAILSEND.get('SMTP_WORKER_EHLO_AS'))

    def send_envelope(self, envelope, attempts=0, priority=60):
        #
        # TODO: set priority using Customer reputation score
        # Best would be to call external modules (like munchers)
        # And keep transactional more prioritized than campaigns
        #
        identifier = envelope.headers.get(
            settings.MAILSEND['X_MESSAGE_ID_HEADER'])

        if self.sandbox:
            self._handle_sandbox(identifier, envelope.recipients[0])
            return
        return self._route(
            identifier, dict(envelope.headers), attempts, priority=priority)

    def send_message(self, identifier, recipient, headers, attempts=0):
        #
        # TODO: set priority using Customer reputation score
        # Best would be to call external modules (like munchers)
        #
        if self.sandbox:
            self._handle_sandbox(identifier, recipient)
            return
        return self._route(identifier, headers, attempts=attempts, priority=50)

    # To be used by Django as a standard email backend
    def send_messages(self, email_messages):
        if not email_messages:
            return
        num_sent = 0
        for message in email_messages:
            if self.send_simple_message(message, system=True):
                num_sent += 1
        return num_sent

    def send_simple_envelope(self, envelope, identifier=None, priority=50):
        if not identifier:
            identifier = envelope.headers.get(
                settings.MAILSEND['X_MESSAGE_ID_HEADER'])
            if identifier is None:
                identifier = mk_base64_uuid()
                envelope.headers.add_header(
                    settings.MAILSEND['X_MESSAGE_ID_HEADER'], identifier)

        raw_mail, created = RawMail.objects.get_or_create(
            content=envelope.message)
        mail = Mail.objects.create(
            identifier=identifier,
            headers=dict(envelope.headers), message=raw_mail,
            sender=envelope.sender, recipient=envelope.recipients[0])
        mailstatus = self.mailstatus_class(
            mail=mail, status=AbstractMailStatus.QUEUED,
            destination_domain=extract_domain(envelope.recipients[0]))

        self.record_status_task(
            mailstatus, identifier,
            settings.MAILSEND.get('SMTP_WORKER_EHLO_AS'))
        self.send_envelope(envelope, priority=priority)

    def send_simple_message(self, message, system=False):
        # Always send system and other simple messages (eg. message previews)
        # before others
        return self.send_simple_envelope(
            message_to_envelope(message), priority=100)



import logging

from celery import current_app as app
from django.core.management.base import BaseCommand

from ...models import Worker

log = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'Remotely shutdown worker.'

    def add_arguments(self, parser):
        parser.add_argument('worker_id', nargs='+', type=str)

    def handle(self, *args, **options):
        for worker_id in options['worker_id']:
            try:
                worker = Worker.objects.get(pk=worker_id)
                self.stdout.write('Saving worker as disabled...')
                worker.enabled = False
                worker.save()
                self.stdout.write('Sending shutdown event to worker...')
                app.control.broadcast('shutdown', destination=(worker.name,))
            except Worker.DoesNotExist:
                self.stdout.write(self.style.WARNING(
                    "* {} doesn't exist (ignored)".format(worker_id)))



from datetime import datetime
from datetime import timedelta

import pytz
from libfaketime import fake_time

from munch_mailsend.models import Mail
from munch_mailsend.models import Worker
from munch_mailsend.models import MailStatus
from munch_mailsend.policies.mx import rate_limit
from munch_mailsend import policies

from . import MailSendTestCase


class RateLimitPolicyTestCase(MailSendTestCase):
    def setUp(self):
        super().setUp()
        self.default_settings = {
            'rate_limit': {
                'domains': [(r'.*', 60)],
                'max_queued': 60 * 15}}

    def test_one_never_send_mail(self):
        worker_01 = Worker.objects.create(
            name='worker_01', ip='10.0.0.1',
            policies_settings=self.default_settings)
        worker_02 = Worker.objects.create(
            name='worker_02', ip='10.0.0.2',
            policies_settings=self.default_settings)

        headers_01 = {'To': 'test+01@example.com'}
        mail_metadata_01 = Mail.objects.create(identifier='0001')

        with fake_time('2015-12-10 12:00:00'):
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_01, source_ip=worker_01.ip)
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_01,
                status=MailStatus.SENDING,
                source_ip=worker_01.ip)
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_01,
                status=MailStatus.DELIVERED,
                source_ip=worker_01.ip)

        with fake_time('2015-12-10 12:00:30'):
            workers = policies.mx.First().apply(headers_01)
            worker_ranking = rate_limit.Policy(
                mail_metadata_01.identifier,
                headers_01, MailStatus).apply(workers)

        self.assertEqual(worker_ranking[0]['score'], 0.2)
        self.assertEqual(worker_ranking[0]['ip'], worker_02.ip)
        self.assertEqual(worker_ranking[1]['score'], 0.1)
        self.assertEqual(worker_ranking[1]['ip'], worker_01.ip)

        best_worker, next_available, *_ = policies.mx.Last().apply(
            worker_ranking)
        self.assertEqual(best_worker.name, worker_02.name)
        self.assertEqual(best_worker.ip, worker_02.ip)

    def test_simple_ordering(self):
        worker_01 = Worker.objects.create(
            name='worker_01', ip='10.0.0.1',
            policies_settings=self.default_settings)
        worker_02 = Worker.objects.create(
            name='worker_02', ip='10.0.0.2',
            policies_settings=self.default_settings)

        headers_01 = {'To': 'test+01@example.com'}
        headers_02 = {'To': 'test+02@example.com'}
        headers_03 = {'To': 'test+03@example.com'}
        headers_04 = {'To': 'test+04@example.com'}
        mail_metadata_01 = Mail.objects.create(identifier='0001')
        mail_metadata_02 = Mail.objects.create(identifier='0002')
        mail_metadata_03 = Mail.objects.create(identifier='0003')
        mail_metadata_04 = Mail.objects.create(identifier='0004')

        # Worker 02 send first mail to test+03@example2.com
        with fake_time('2015-12-10 12:00:00'):
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_01, source_ip=worker_02.ip)
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_01,
                status=MailStatus.DELIVERED,
                source_ip=worker_02.ip)
        # Worker 01 send mail to test+01@example.com
        with fake_time('2015-12-10 12:00:10'):
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_02, source_ip=worker_01.ip)
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_02,
                status=MailStatus.DELIVERED,
                source_ip=worker_01.ip)
        # Worker 02 send mail to test+02@example.com
        with fake_time('2015-12-10 12:00:20'):
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_03, source_ip=worker_02.ip)
            MailStatus.objects.create(
                destination_domain='example.com',
                status=MailStatus.SENDING,
                mail=mail_metadata_03, source_ip=worker_02.ip)

        with fake_time('2015-12-10 12:00:15'):
            # Test for test+01@example.com
            workers = policies.mx.First().apply(headers_01)
            worker_ranking = rate_limit.Policy(
                mail_metadata_04.identifier,
                headers_04, MailStatus).apply(workers)

            self.assertEqual(worker_ranking[0]['score'], 0.2)
            self.assertEqual(worker_ranking[0]['ip'], worker_01.ip)
            self.assertEqual(worker_ranking[1]['score'], 0.1)
            self.assertEqual(worker_ranking[1]['ip'], worker_02.ip)

            best_worker, next_available, *_ = policies.mx.Last().apply(
                worker_ranking)
            self.assertEqual(best_worker.name, worker_01.name)
            self.assertEqual(best_worker.ip, worker_01.ip)

            # Test for test+02@example.com
            # Must be the same as test+02@example.com
            workers = policies.mx.First().apply(headers_02)
            worker_ranking = rate_limit.Policy(
                mail_metadata_02.identifier,
                headers_02, MailStatus).apply(workers)

            self.assertEqual(worker_ranking[0]['score'], 0.2)
            self.assertEqual(worker_ranking[0]['ip'], worker_01.ip)
            self.assertEqual(worker_ranking[1]['score'], 0.1)
            self.assertEqual(worker_ranking[1]['ip'], worker_02.ip)

            best_worker, next_available, *_ = policies.mx.Last().apply(
                worker_ranking)
            self.assertEqual(best_worker.name, worker_01.name)
            self.assertEqual(best_worker.ip, worker_01.ip)

            # Test for test+03@example2.com
            workers = policies.mx.First().apply(headers_03)
            worker_ranking = rate_limit.Policy(
                mail_metadata_03.identifier, headers_03, MailStatus).apply(
                    workers)

            self.assertEqual(worker_ranking[0]['score'], 0.2)
            self.assertEqual(worker_ranking[0]['ip'], worker_01.ip)
            self.assertEqual(worker_ranking[1]['score'], 0.1)
            self.assertEqual(worker_ranking[1]['ip'], worker_02.ip)

            best_worker, next_available, *_ = policies.mx.Last().apply(
                worker_ranking)
            self.assertEqual(best_worker.name, worker_01.name)
            self.assertEqual(best_worker.ip, worker_01.ip)

    def test_interval_between_sending(self):
        worker = Worker.objects.create(
            name='worker', ip='10.0.0.1',
            policies_settings=self.default_settings)

        with fake_time('2015-12-10 12:00:00'):
            mail_01 = Mail.objects.create(identifier='0001')
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_01, source_ip=worker.ip)
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_01, status=MailStatus.SENDING, source_ip=worker.ip)

        with fake_time('2015-12-10 12:05:00'):
            mail_02 = Mail.objects.create(identifier='0002')
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_02, source_ip=worker.ip)
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_02, status=MailStatus.SENDING, source_ip=worker.ip)

        with fake_time('2015-12-10 12:01:00'):
            mail_03 = Mail.objects.create(identifier='0003')
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_03, source_ip=worker.ip)
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_03, status=MailStatus.SENDING, source_ip=worker.ip)

            workers = policies.mx.First().apply({'To': 'test@example.com'})
            worker_ranking = rate_limit.Policy(
                mail_03.identifier,
                {'To': 'test@example.com'}, MailStatus).apply(workers)

            self.assertEqual(len(worker_ranking), 1)

            best_worker, next_available, *_ = policies.mx.Last().apply(
                worker_ranking)

            self.assertEqual(best_worker.name, worker.name)
            self.assertEqual(next_available, datetime.now(
                pytz.UTC) + timedelta(minutes=1))

    def test_setting_limiting(self):
        worker_01 = Worker.objects.create(
            name='worker_01', ip='10.0.0.1',
            policies_settings=self.default_settings)
        worker_02 = Worker.objects.create(
            name='worker_02', ip='10.0.0.2',
            policies_settings=self.default_settings)

        headers_01 = {'To': 'test+01@example.com'}
        mail_metadata_01 = Mail.objects.create(identifier='0001')

        with fake_time('2015-12-10 12:00:00'):
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_01, source_ip=worker_01.ip)
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_01,
                status=MailStatus.SENDING,
                source_ip=worker_01.ip)
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_01,
                status=MailStatus.DELIVERED,
                source_ip=worker_01.ip)

        with fake_time('2015-12-10 12:00:30'):
            workers = policies.mx.First().apply(headers_01)
            worker_ranking = rate_limit.Policy(
                mail_metadata_01.identifier,
                headers_01, MailStatus).apply(workers)

        self.assertEqual(len(worker_ranking), 2)

        self.assertEqual(worker_ranking[0]['score'], 0.2)
        self.assertEqual(worker_ranking[0]['ip'], worker_02.ip)
        self.assertEqual(worker_ranking[0]['name'], worker_02.name)

        best_worker, next_available, *_ = policies.mx.Last().apply(
            worker_ranking)
        self.assertEqual(best_worker.name, worker_02.name)
        self.assertEqual(best_worker.ip, worker_02.ip)

    def test_setting_domain_catched(self):
        worker_01 = Worker.objects.create(
            name='worker_01', ip='10.0.0.1', policies_settings={
                'rate_limit': {
                    'domains': [
                        (r'example.com', 60 * 5),
                        (r'.*', 60)],
                    'max_queued': 60 * 15}})

        headers_01 = {'To': 'test+01@example.com'}
        headers_02 = {'To': 'test+02@example.com'}
        mail_01 = Mail.objects.create(identifier='0001')
        mail_02 = Mail.objects.create(identifier='0002')

        with fake_time('2015-12-10 12:00:00'):
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_01, source_ip=worker_01.ip)
            MailStatus.objects.create(
                destination_domain='example.com', mail=mail_01,
                status=MailStatus.SENDING, source_ip=worker_01.ip)
            MailStatus.objects.create(
                destination_domain='example.com', mail=mail_01,
                status=MailStatus.DELIVERED, source_ip=worker_01.ip)

        with fake_time('2015-12-10 12:00:00'):
            workers = policies.mx.First().apply(headers_01)
            worker_ranking = rate_limit.Policy(
                mail_01.identifier, headers_01, MailStatus).apply(workers)

            self.assertEqual(len(worker_ranking), 1)

            best_worker, next_available, *_ = policies.mx.Last().apply(
                worker_ranking)
            self.assertEqual(best_worker, worker_01)
            self.assertEqual(next_available, datetime.now(
                pytz.UTC) + timedelta(minutes=5))

        with fake_time('2015-12-10 12:05:00'):
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_02, source_ip=worker_01.ip)
            MailStatus.objects.create(
                destination_domain='example.com', mail=mail_02,
                status=MailStatus.DELIVERED, source_ip=worker_01.ip)

        with fake_time('2015-12-10 12:05:30'):
            workers = policies.mx.First().apply(headers_02)
            worker_ranking = rate_limit.Policy(
                mail_02.identifier, headers_02, MailStatus).apply(workers)

        self.assertEqual(len(worker_ranking), 1)
        self.assertEqual(worker_ranking[0]['score'], 0.1)
        self.assertEqual(worker_ranking[0]['ip'], worker_01.ip)
        self.assertEqual(worker_ranking[0]['name'], worker_01.name)

        best_worker, next_available, *_ = policies.mx.Last().apply(
            worker_ranking)
        self.assertEqual(best_worker.name, worker_01.name)
        self.assertEqual(best_worker.ip, worker_01.ip)



from munch_mailsend.models import Worker
from munch_mailsend.policies.mx import First

from . import MailSendTestCase


class WorkersTestCase(MailSendTestCase):
    def test_retrieve_workers(self):
        Worker.objects.create(name='worker_01', ip='10.0.0.1')
        Worker.objects.create(name='worker_02', ip='10.0.0.2')

        workers = First().apply({})
        self.assertEqual(len(workers), 2)
        Worker.objects.clear_cache()
        workers = First().apply({})
        self.assertEqual(len(workers), 2)



import random
from datetime import date
from datetime import timedelta

from django.conf import settings
from django_redis import get_redis_connection

from . import CACHE_PREFIX
from . import CACHE_TIMEOUT
from . import WorkerPolicyBase

conn = get_redis_connection('default')


class Policy(WorkerPolicyBase):
    """
        Warm up

        Prioritize: "equal", "warmest", "coldest"

        {
            'prioritize': 'equal',
            'domain_warm_up': {
                'matrix': [50, 100, 300, 500, 1000],
                'goal': 500,
                'max_tolerance': 10,
                'step_tolerance': 10,
                'days_watched': 10,
            },
            'ip_warm_up': {
                'matrix': [50, 100, 300, 500, 1000],
                'goal': 500,
                'max_tolerance': 10,
                'step_tolerance': 10,
                'enabled': False,
                'days_watched': 10,
            }
        }
    """
    def get_specific_settings(self, domain, policy_settings):
        if self.get_domain_group(domain):
            return policy_settings.get('domain_warm_up', {})
        elif policy_settings.get('ip_warm_up', {}).get('enabled'):
            return policy_settings.get('ip_warm_up', {})
        return {}

    def apply(self, workers):
        today = date.today()
        domain = self.get_domain(self.headers.get('To'))
        mailstatus_cache_timeout = int(settings.MAILSEND[
            'MAILSTATUS_CACHE_TIMEOUT'] / (60 * 60 * 24))

        for worker in workers:
            policy_settings = self.get_specific_settings(
                domain, self.get_settings(worker))

            domain_group = self.get_domain_group(domain)
            mail_statuses_kwargs = {'source_ip': worker.get('ip')}
            if domain_group:
                self.logger.debug(
                    '[{}] [worker:{}] This envelope will have a '
                    'warm-up just for its domain group ({})'.format(
                        self.identifier, worker.get('ip'), domain_group))
                mail_statuses_kwargs.update(
                    {'destination_domain': self.get_shared_domains(
                        domain_group)})
            elif policy_settings.get('enabled', False):
                self.logger.debug(
                    '[{}] [worker:{}] Applying ip warm-up...'.format(
                        self.identifier, worker.get('ip')))
            else:
                self.logger.debug(
                    '[{}] [worker:{}] This envelope will not have any '
                    'warm-up share with these domains ({}). '
                    'Missing percent set at 100%.'.format(
                        self.identifier, worker.get('ip'), domain))
                worker['_warm_up_step'] = 0
                worker['_warm_up_missing_percent'] = 100
                continue

            days_watched = int(policy_settings.get(
                'days_watched', mailstatus_cache_timeout))

            if days_watched > mailstatus_cache_timeout:
                self.logger.warning(
                    '[{}] [worker:{}] "days_watched" is higher than '
                    '"MAILSEND[\'MAILSTATUS_CACHE_TIMEOUT\']". Fallback to '
                    'this value.'.format(self.identifier, worker.get('ip')))
                days_watched = mailstatus_cache_timeout

            # Retrieve step from cache
            step, remains = self.get_step(
                worker.get('ip'), domain_group=domain_group)
            # If step return is None, lets search and cache it
            if step is None:
                self.logger.debug(
                    '[{}] [worker:{}] Building cache for step '
                    'and remains...'.format(self.identifier, worker.get('ip')))
                step_tolerance = int(policy_settings.get('step_tolerance'))
                goal = int(policy_settings.get('goal'))
                matrix = policy_settings.get('matrix', [])
                step = self.search_step(
                    worker=worker, days_watched=days_watched, goal=goal,
                    step_tolerance=step_tolerance, matrix=matrix, today=today,
                    mail_statuses_kwargs=mail_statuses_kwargs)
                remains = step + int(policy_settings.get(
                    'max_tolerance', 0) * step / 100)
                self.set_step(
                    worker.get('ip'), step, remains, domain_group=domain_group)

            missing_percent = 100 - int((step - remains) * 100 / step)
            self.logger.debug(
                '[{}] [worker:{}] Remains for today is {} for {} '
                'step (missing_percent:{})'.format(
                    self.identifier, worker.get('ip'),
                    remains, step, missing_percent))
            worker['_warm_up_step'] = step
            worker['_warm_up_missing_percent'] = missing_percent

        return self.apply_prioritize(workers, domain)

    def apply_prioritize(self, workers, domain):
        prioritize = settings.MAILSEND.get(
            'WORKER_POLICIES_SETTINGS', {}).get('warm_up', {}).get(
                'prioritize', 'equal').lower()

        if prioritize == 'warmest':
            workers = sorted(
                workers, key=lambda k: (k['_warm_up_step'], random.random()))
        elif prioritize == 'coldest':
            workers = sorted(
                workers, reverse=True,
                key=lambda k: (k['_warm_up_step'], random.random()))

        ranked_workers = []
        for i, worker in enumerate(workers):
            if worker.get('_warm_up_missing_percent') <= 0:
                self.logger.debug(
                    "[{}] [worker:{}] Can't keep this worker because it reach"
                    " its step + max_tolerance: {}".format(
                        self.identifier, worker.get('ip'), 100 - worker.get(
                            '_warm_up_missing_percent')))
                continue
            percent = worker.get('_warm_up_missing_percent')
            if prioritize == 'equal':
                worker['score'] += round(percent * 0.01 / len(workers), 2)
            elif prioritize == 'warmest':
                worker['score'] += round(
                    percent * 0.01 / len(workers), 2) + 0.1 * i
            elif prioritize == 'coldest':
                worker['score'] += round(
                    percent * 0.01 / len(workers), 2) + 0.1 * i
            ranked_workers.append(worker)

        ranked_workers = sorted(ranked_workers, key=lambda k: k['score'])
        for worker in ranked_workers:
            worker.pop('_warm_up_step')
            worker.pop('_warm_up_missing_percent')
        return ranked_workers

    def search_step(
            self, worker, days_watched, matrix, goal,
            step_tolerance, today, mail_statuses_kwargs):
        step = 0

        for days in range(days_watched):
            days += 1

            day_to_watch = today - timedelta(days=days)
            mail_statuses_kwargs.update({'creation_date': day_to_watch})
            latest_counter = self.get_counter(**mail_statuses_kwargs)

            if not latest_counter:
                if matrix[0] > step:
                    step = matrix[0]
                    self.logger.debug(
                        '[{}] [worker:{}] Not statuses for today - {} '
                        'day(s). Start matrix at first '
                        'step, which is {}.'.format(
                            self.identifier, worker.get('ip'), days, step))
            else:
                self.logger.debug(
                    '[{}] [worker:{}] Today - {} day(s) '
                    'statuses count: {}.'.format(
                        self.identifier, worker.get('ip'),
                        days, latest_counter))
                for i, _step in enumerate(matrix):
                    if len(matrix) > i + 1:
                        next_step = matrix[i + 1]
                    else:
                        next_step = matrix[i]

                    up_counter = int(
                        next_step - next_step * step_tolerance / 100)
                    down_counter = int(
                        step - step * step_tolerance / 100)
                    if down_counter <= latest_counter < up_counter:
                        if int(matrix[i + 1]) > step:
                            self.logger.debug(
                                '[{}] [worker:{}] Found a new higher'
                                ' step at {} from {}.'.format(
                                    self.identifier, worker.get('ip'),
                                    int(matrix[i + 1]), step))
                            step = int(matrix[i + 1])
                            break

            if step > goal:
                self.logger.debug(
                    '[{}] [worker:{}] Step is greater than goal. Set step '
                    'to goal ({} => {})'.format(
                        self.identifier, worker.get('ip'), step, goal))
                step = goal

            return step

    @staticmethod
    def get_domain_group(domain):
        for domain_group, domains in settings.MAILSEND.get(
                'WARM_UP_DOMAINS', {}).items():
            if domain in domains:
                return domain_group

    @staticmethod
    def get_shared_domains(domain_group):
        return settings.MAILSEND.get(
            'WARM_UP_DOMAINS', {}).get(domain_group, [])

    @classmethod
    def update_counter(
            cls, source_ip, destination_domain, value, creation_date=None):
        if creation_date:
            creation_date = creation_date.strftime('%Y-%m-%d')
        else:
            creation_date = date.today().strftime('%Y-%m-%d')
        domain_group = cls.get_domain_group(destination_domain)
        key_base = '{}:warm_up:remains:{}:{}'.format(
            CACHE_PREFIX, creation_date, source_ip)
        if value == 1:
            if conn.get(key_base) is None:
                conn.set(key_base, 1, CACHE_TIMEOUT)
            else:
                conn.incr(key_base)
            if domain_group:
                key_base = '{}:{}'.format(key_base, domain_group)
                if conn.get(key_base) is None:
                    conn.set(key_base, 1, CACHE_TIMEOUT)
                else:
                    conn.incr(key_base)
        elif value == -1:
            if conn.get(key_base) is None:
                conn.set(key_base, -1, CACHE_TIMEOUT)
            else:
                conn.decr(key_base)
            if domain_group:
                key_base = '{}:{}'.format(key_base, domain_group)
                if conn.get(key_base) is None:
                    conn.set(key_base, -1, CACHE_TIMEOUT)
                else:
                    conn.decr(key_base)

    @staticmethod
    def set_step(
            source_ip, step, remains, domain_group=None):
        creation_date = date.today().strftime('%Y-%m-%d')
        ########
        # Step #
        ########
        key_base = '{}:warm_up:step:{}:{}'.format(
            CACHE_PREFIX, creation_date, source_ip)
        if not domain_group:
            conn.set(key_base, step, CACHE_TIMEOUT)
        else:
            conn.set(key_base + ':' + domain_group, step, CACHE_TIMEOUT)
        ###########
        # Remains #
        ###########
        key_base = '{}:warm_up:remains:{}:{}'.format(
            CACHE_PREFIX, creation_date, source_ip)
        if not domain_group:
            if conn.get(key_base) is None:
                conn.set(key_base, remains, CACHE_TIMEOUT)
            else:
                conn.incr(key_base, remains)
        else:
            key_base = key_base + ':' + domain_group
            if conn.get(key_base) is None:
                conn.set(key_base, remains, CACHE_TIMEOUT)
            else:
                conn.incr(key_base, remains)

    @staticmethod
    def get_step(source_ip, domain_group=None):
        step, remains = None, None
        creation_date = date.today().strftime('%Y-%m-%d')
        ########
        # Step #
        ########
        key_base = '{}:warm_up:step:{}:{}'.format(
            CACHE_PREFIX, creation_date, source_ip)
        if not domain_group:
            step = conn.get(key_base)
        else:
            step = conn.get(key_base + ':' + domain_group)
        ###########
        # Remains #
        ###########
        key_base = '{}:warm_up:remains:{}:{}'.format(
            CACHE_PREFIX, creation_date, source_ip)
        if not domain_group:
            remains = conn.get(key_base)
        else:
            remains = conn.get(key_base + ':' + domain_group)

        try:
            step = int(step)
        except (ValueError, TypeError):
            step = None
        try:
            remains = int(remains)
        except (ValueError, TypeError):
            remains = None

        return step, remains

    @staticmethod
    def get_counter(creation_date, source_ip, destination_domain=None):
        key_base = '{}:warm_up:counter:{}:{}'.format(
            CACHE_PREFIX,
            creation_date.strftime('%Y-%m-%d'), source_ip or '*')
        if not destination_domain:
            try:
                return int(conn.get(key_base))
            except (ValueError, TypeError):
                return 0
        if not isinstance(destination_domain, list) and not isinstance(
                destination_domain, tuple):
            destination_domain = (destination_domain, )
        counter = 0
        for domain in destination_domain:
            for key in conn.scan_iter('{}:{}'.format(key_base, domain)):
                try:
                    counter += int(conn.get(key))
                except (ValueError, TypeError):
                    pass
        return counter

    ###########
    # Signals #
    ###########

    @classmethod
    def mailstatus_pre_save(cls, instance, manager):
        if instance.status in [instance.SENDING]:
            cls.update_counter(
                instance.source_ip, instance.destination_domain,
                -1, creation_date=instance.creation_date)
        if instance.status in [instance.DELAYED]:
            cls.update_counter(
                instance.source_ip, instance.destination_domain,
                +1, creation_date=instance.creation_date)
        if instance.status in [instance.DELIVERED, instance.BOUNCED]:
            # Key example which is a counter:
            # ms:warm_up:1990-01-01:127.0.0.1:postfix.example.com
            key = '{}:warm_up:counter:{}:{}:{}'.format(
                CACHE_PREFIX,
                instance.creation_date.strftime('%Y-%m-%d'),
                instance.source_ip, instance.destination_domain)
            key_worker = '{}:warm_up:counter:{}:{}'.format(
                CACHE_PREFIX,
                instance.creation_date.strftime('%Y-%m-%d'),
                instance.source_ip)
            if conn.get(key) is None:
                conn.set(key, 0, CACHE_TIMEOUT)
            if conn.get(key_worker) is None:
                conn.set(key_worker, 0, CACHE_TIMEOUT)
            conn.incr(key)
            conn.incr(key_worker)



import re
import random
from datetime import datetime
from datetime import timedelta

import pytz
from django_redis import get_redis_connection

from . import CACHE_PREFIX
from . import WorkerPolicyBase

# from celery.contrib import rdb

conn = get_redis_connection('default')


class Policy(WorkerPolicyBase):
    """
        Compute next_available based on latest sending status

        # Example of settings
        priotitize options: `earlier`, `equal`
        {
            'domains': [
                (re.compile(r'.*oasiswork.fr'), 60 * 1),
                (re.compile(r'.*'), 60 * 5)],
            'max_queued': 60 * 30,
            'prioritize': 'earlier'
        }
    """
    def apply(self, workers):
        domain = self.get_domain(self.headers.get('To'))

        now = self.now()
        old = self.old()
        # Taking not_before or we are assuming it's now if not specified
        not_before = self.not_before or now
        for worker in workers:
            prioritize = self.get_settings(worker).get('prioritize', 'earlier')

            domain_limit = 0
            for domain_settings in self.get_settings(
                    worker).get('domains', []):
                if re.match(domain_settings[0], domain):
                    domain_limit = domain_settings[1]
                    self.logger.debug(
                        '[{}] [worker:{}] Domain rate limiting detected at {} '
                        'second(s) for {} (domain:{})'.format(
                            self.identifier, worker.get('ip'),
                            domain_limit, domain, domain_settings[0]))
                    break

            statuses = sorted(self.get_statuses(
                source_ip=worker.get('ip'),
                destination_domain=domain,
                creation_date=now - timedelta(seconds=domain_limit)),
                key=lambda ms: ms.get('creation_date'), reverse=False)
            next_available = None
            # If there is no statuses, we can set next_available at now
            if not statuses:
                next_available = not_before
                self.logger.debug(
                    '[{}] [worker:{}] No previous sending statuses, then '
                    'next_available is now or not_before ({}).'.format(
                        self.identifier, worker.get('ip'),
                        next_available.astimezone()))

            # First check if we can insert before the first scheduled mail
            if statuses:
                status_date = statuses[0].get('creation_date') or old
                if now + timedelta(seconds=domain_limit * 2) < status_date:
                    next_available = now + timedelta(seconds=domain_limit)
                # If next_available is before the not_before constraint
                # Unset next_available and let's continue searching
                if next_available and next_available < not_before:
                    self.logger.debug(
                        '[{}] [worker:{}] Envelope has a not_before ({}) '
                        'and next_available ({}) is too early, '
                        'searching another one...'.format(
                            self.identifier, worker.get('ip'),
                            not_before.astimezone(),
                            next_available.astimezone()))
                    next_available = None

            if not next_available:
                # Keep searching for a next_available after the first
                # scheduled mail
                for i, status in enumerate(statuses):
                    next_status = None
                    status_date = status.get('creation_date') or old
                    # Check if there is another sending status after this one
                    if len(statuses) > i + 1:
                        next_status = statuses[i + 1].get('creation_date')
                    # If it's the only one status or we can schedule
                    # a sending between this and next one
                    if not next_status or status_date + timedelta(
                            seconds=domain_limit * 2) < next_status:
                        # Then compute next available
                        next_available = status_date + timedelta(
                            seconds=domain_limit)
                        # If next_available is before the not_before constraint
                        # Unset next_available and let's continue searching
                        if next_available < not_before:
                            self.logger.debug(
                                '[{}] [worker:{}] Envelope has a not_before '
                                '({}) and next_available ({}) is too early, '
                                'searching another one...'.format(
                                    self.identifier, worker.get('ip'),
                                    not_before.astimezone(),
                                    next_available.astimezone()))
                            next_available = None
                        # Else, break for loop with computed next_available
                        else:
                            self.logger.debug(
                                '[{}] [worker:{}] Potential '
                                'next_available found at {}'.format(
                                    self.identifier,
                                    worker.get('ip'),
                                    next_available.astimezone()))
                            break

            # If we don't find any next_available
            # (because of no statuses for example) or next_available is in past
            # Then we set next_available at not_before which
            # is now if not set.
            if not next_available or next_available < now:
                next_available = not_before
            # If next_available is after the value choosen by a previous policy
            # We override it.
            if next_available > worker.get('next_available', now):
                worker['next_available'] = next_available
            self.logger.debug(
                '[{}] [worker:{}] Final next_available found at {}.'.format(
                    self.identifier, worker.get('ip'),
                    worker.get('next_available').astimezone()))

        # Then order available workers based on next_available
        def get_next_available(worker):
            if worker.get('next_available'):
                return (worker.get('next_available'), random.random())
            return datetime(year=1970, month=1, day=1).replace(
                tzinfo=pytz.utc), random.random()

        workers = sorted(workers, key=get_next_available)
        ranked_workers = []
        for index, worker in enumerate(workers):
            # Discard max_queued
            max_queued = int(self.get_settings(worker).get('max_queued', 30))
            max_queued_datetime = now + timedelta(seconds=max_queued)
            if worker.get('next_available') > max_queued_datetime:
                self.logger.debug(
                    '[{}] [worker:{}] Next available is '
                    'too far (max_queue:{}) to be scheduled '
                    'for this worker (next_available:{})'.format(
                        self.identifier, worker.get('ip'),
                        max_queued_datetime,
                        worker.get('next_available').astimezone()))
                continue
            if prioritize == 'earlier':
                worker['score'] += round((len(workers) - index) * 0.1, 2)
            ranked_workers.append(worker)
        return ranked_workers

    @staticmethod
    def get_statuses(source_ip, destination_domain, creation_date):
        statuses = []
        for status in conn.zrangebyscore(
                '{}:rate_limit:{}:{}'.format(
                    CACHE_PREFIX, source_ip, destination_domain),
                creation_date.timestamp(), '+inf'):
            splitted_status = status.decode('utf-8').split(':')
            statuses.append({
                'identifier': splitted_status[0],
                'creation_date': datetime.fromtimestamp(float(
                    splitted_status[1]), pytz.utc)})
        return statuses

    ###########
    # Signals #
    ###########

    @staticmethod
    def mailstatus_pre_save(instance, manager):
        if instance.status in [instance.SENDING]:
            conn.zadd(
                '{}:rate_limit:{}:{}'.format(
                    CACHE_PREFIX, instance.source_ip,
                    instance.destination_domain),
                instance.creation_date.timestamp(),
                '{}:{}'.format(
                    instance.mail.identifier,
                    instance.creation_date.timestamp()))



from django.test import override_settings

from munch_mailsend.relay import MxSmtpRelay

from . import MailSendTestCase


class RelayTestCase(MailSendTestCase):
    @override_settings(MAILSEND={'SMTP_WORKER_EHLO_AS': None})
    def test_default_relay_ehlo(self):
        relay = MxSmtpRelay()
        self.assertEqual(relay._client_kwargs['ehlo_as'], None)

    @override_settings(MAILSEND={'SMTP_WORKER_EHLO_AS': 'test12'})
    def test_override_relay_ehlo(self):
        relay = MxSmtpRelay()
        self.assertEqual(relay._client_kwargs['ehlo_as'], 'test12')



from email.utils import parseaddr

from django.conf import settings
from slimta.policy import RelayPolicy

from munch.core.mail.utils import extract_domain

from ...utils.dkim import sign
from ...utils.dkim import DKIMException

# Headers must be Relaxed-Canonicalized
# because dkim doesn't do it automatically
DEFAULT_HEADERS_TO_SIGN = [
    'From', 'Subject', 'To', 'Date', 'Message-ID', 'Content-Type',
    'MIME-Version', settings.MAILSEND['X_USER_ID_HEADER'],
    settings.MAILSEND['X_MESSAGE_ID_HEADER']]


class Sign(RelayPolicy):
    def apply(self, envelope):

        # Copy without reference to avoid modifying default headers
        inc_headers = DEFAULT_HEADERS_TO_SIGN[:]
        # Add headers defined in settings
        for h in settings.MAILSEND.get('DKIM_EXTRA_SIGN_HEADERS', []):
            inc_headers.append(h)
        # Add extra headers to sign list if they are present
        # if not, we don't want to prevent them from being added later
        for h in ['Reply-To', 'List-ID', 'List-Unsubscribe', 'Sender']:
            if h in envelope.headers:
                inc_headers.append(h)
        encoded_inc_headers = [h.encode('utf-8') for h in inc_headers]

        identity = parseaddr(envelope.headers['From'])[1]
        domain = extract_domain(identity).encode('utf-8')

        try:
            headers_data, message_data = envelope.flatten()
            dkim_header = sign(
                headers_data + message_data,
                settings.MAILSEND.get('DKIM_SELECTOR', '').encode('utf-8'),
                domain,
                settings.MAILSEND.get('DKIM_PRIVATE_KEY', '').encode('utf-8'),
                identity=identity.encode('utf-8'), length=False,
                include_headers=encoded_inc_headers)
            dkim_header = dkim_header.decode('utf-8')
            if dkim_header.startswith('DKIM-Signature: '):
                dkim_header = dkim_header[16:]
                envelope.headers['DKIM-Signature'] = dkim_header
        except DKIMException:
            raise



from django.conf import settings

from .policies import run_policies


def pre_save_mailstatus(sender, instance, raw, **kwargs):
    # Default values
    if not instance.pk and not raw:
        if not instance.source_ip:
            instance.source_ip = settings.MAILSEND.get('SMTP_WORKER_SRC_ADDR')
    # Policies
    if not instance.pk:
        run_policies(instance, 'mailstatus_pre_save')


def post_save_mailstatus(sender, instance, created, raw, **kwargs):
    # Policies
    if created:
        run_policies(instance, 'mailstatus_post_save')



from django.test import TestCase
from django.conf import settings
from libfaketime import reexec_if_needed
from django_redis import get_redis_connection

reexec_if_needed()


class MailSendTestCase(TestCase):
    def tearDown(self):
        conn = get_redis_connection()
        for key in conn.scan_iter('{}:*'.format(
                settings.MAILSEND['CACHE_PREFIX'])):
            conn.delete(key)



from django.conf import settings

from . import WorkerPolicyBase


class Policy(WorkerPolicyBase):
    """
        settings = {"pools": ['default']}
    """
    def apply(self, workers):
        available_workers = []
        x_pool_header = settings.MAILSEND['X_POOL_HEADER']
        pool = self.headers.get(x_pool_header, '').strip().lower()
        if not pool:
            pool = 'default'
            self.logger.debug(
                '[{}] No "{}" header found. Using: {}'.format(
                    self.identifier, x_pool_header, pool))
        else:
            self.logger.debug(
                '[{}] Found "{}" header with: {}'.format(
                    self.identifier, x_pool_header, pool))

        for worker in workers:
            worker_pools = self.get_settings(worker).get('pools', ['default'])

            if pool in worker_pools:
                self.logger.debug(
                    '[{}] [worker:{}] Pool: "{}" matched in {}'.format(
                        self.identifier, worker.get('ip'),
                        pool, worker_pools))
                available_workers.append(worker)
            else:
                self.logger.debug(
                    '[{}] [worker:{}] No pool matched for {} in {}.'.format(
                        self.identifier, worker.get('ip'), pool, worker_pools))
        return available_workers



from libfaketime import fake_time

from munch_mailsend.models import Mail
from munch_mailsend.models import MailStatus
from munch_mailsend import policies
from munch_mailsend.models import Worker
from munch_mailsend.policies.mx import pool

from . import MailSendTestCase


class PoolPolicyTestCase(MailSendTestCase):
    def setUp(self):
        super().setUp()
        self.default_settings = {
            'pool': {
                'X_POOL_HEADER': 'X-MAILSEND-Pool',
                'pools': ['default']
            }
        }

    def test_pool(self):
        worker_01 = Worker.objects.create(
            name='worker_01', ip='10.0.0.1',
            policies_settings=self.default_settings)
        Worker.objects.create(
            name='worker_02', ip='10.0.0.2',
            policies_settings={
                'pool': {
                    'pools': ['jambon'],
                    'X_POOL_HEADER': 'X-MAILSEND-Pool'}})
        mail_metadata_01 = Mail.objects.create(identifier='0001')

        with fake_time('2015-12-10 12:00:00'):
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_01, source_ip=worker_01.ip)

        with fake_time('2015-12-10 12:00:05'):
            workers = policies.mx.First().apply({'To': 'test+01@example.com'})
            worker_ranking = pool.Policy(
                mail_metadata_01.identifier,
                {'To': 'test+01@example.com'},
                MailStatus).apply(workers)

            self.assertEqual(len(worker_ranking), 1)

            best_worker, next_available, score, *_ = policies.mx.Last().apply(
                worker_ranking)

            self.assertEqual(best_worker.name, worker_01.name)
            self.assertEqual(best_worker.ip, worker_01.ip)
            self.assertEqual(score, 0.0)

    def test_pool_empty_list_is_default(self):
        worker_01 = Worker.objects.create(
            name='worker_01', ip='10.0.0.1',
            policies_settings={'pool': {'X_POOL_HEADER': 'X-MAILSEND-Pool'}})
        Worker.objects.create(
            name='worker_02', ip='10.0.0.2',
            policies_settings={
                'pool': {
                    'pools': ['jambon'],
                    'X_POOL_HEADER': 'X-MAILSEND-Pool'}})
        mail_metadata_01 = Mail.objects.create(identifier='0001')

        with fake_time('2015-12-10 12:00:00'):
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_01, source_ip=worker_01.ip)

        with fake_time('2015-12-10 12:00:05'):
            workers = policies.mx.First().apply({'To': 'test+01@example.com'})
            worker_ranking = pool.Policy(
                mail_metadata_01.identifier,
                {'To': 'test+01@example.com'},
                MailStatus).apply(workers)

            self.assertEqual(len(worker_ranking), 1)

            best_worker, next_available, score, *_ = policies.mx.Last().apply(
                worker_ranking)

            self.assertEqual(best_worker.name, worker_01.name)
            self.assertEqual(best_worker.ip, worker_01.ip)
            self.assertEqual(score, 0.0)

    def test_pool_no_match(self):
        worker_01 = Worker.objects.create(
            name='worker_01', ip='10.0.0.1',
            policies_settings={
                'pool': {'X_POOL_HEADER': 'X-MAILSEND-Pool', 'pools': [
                    'jambon']}})
        Worker.objects.create(
            name='worker_02', ip='10.0.0.2',
            policies_settings={
                'pool': {'X_POOL_HEADER': 'X-MAILSEND-Pool', 'pools': [
                    'jambon']}})
        mail_metadata_01 = Mail.objects.create(identifier='0001')

        with fake_time('2015-12-10 12:00:00'):
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_01, source_ip=worker_01.ip)

        with fake_time('2015-12-10 12:00:05'):
            workers = policies.mx.First().apply({'To': 'test+01@example.com'})
            worker_ranking = pool.Policy(
                mail_metadata_01.identifier,
                {'To': 'test+01@example.com'},
                MailStatus).apply(workers)

            self.assertEqual(len(worker_ranking), 0)

            best_worker, next_available, score, *_ = policies.mx.Last().apply(
                worker_ranking)

            self.assertIsNone(best_worker)

    def test_pool_match_kk(self):
        worker_01 = Worker.objects.create(
            name='worker_01', ip='10.0.0.1',
            policies_settings={
                'pool': {'X_POOL_HEADER': 'X-MAILSEND-Pool', 'pools': [
                    'jambon']}})
        Worker.objects.create(
            name='worker_02', ip='10.0.0.2',
            policies_settings={
                'pool': {
                    'pools': ['foo'],
                    'X_POOL_HEADER': 'X-MAILSEND-Pool'}})
        mail_metadata_01 = Mail.objects.create(
            identifier='0001', headers={'X-MAILSEND-Pool': 'jambon'})

        with fake_time('2015-12-10 12:00:00'):
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_01, source_ip=worker_01.ip)

        with fake_time('2015-12-10 12:00:05'):
            workers = policies.mx.First().apply({'To': 'test+01@example.com'})
            worker_ranking = pool.Policy(
                mail_metadata_01.identifier,
                {'To': 'test+01@example.com', 'X-MAILSEND-Pool': 'jambon'},
                MailStatus).apply(workers)

            self.assertEqual(len(worker_ranking), 1)

            best_worker, next_available, score, *_ = policies.mx.Last().apply(
                worker_ranking)

            self.assertEqual(best_worker.name, worker_01.name)
            self.assertEqual(best_worker.ip, worker_01.ip)
            self.assertEqual(score, 0.0)

    def test_pool_match_insensitive(self):
        worker_01 = Worker.objects.create(
            name='worker_01', ip='10.0.0.1',
            policies_settings={
                'pool': {'X_POOL_HEADER': 'X-MAILSEND-Pool', 'pools': [
                    'jambon']}})
        Worker.objects.create(
            name='worker_02', ip='10.0.0.2',
            policies_settings={
                'pool': {
                    'pools': ['foo'],
                    'X_POOL_HEADER': 'X-MAILSEND-Pool'}})
        mail_metadata_01 = Mail.objects.create(
            identifier='0001', headers={'X-MAILSEND-Pool': 'JAmbon'})

        with fake_time('2015-12-10 12:00:00'):
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_01, source_ip=worker_01.ip)

        with fake_time('2015-12-10 12:00:05'):
            workers = policies.mx.First().apply({'To': 'test+01@example.com'})
            worker_ranking = pool.Policy(
                mail_metadata_01.identifier,
                {'To': 'test+01@example.com', 'X-MAILSEND-Pool': 'JAmbon'},
                MailStatus).apply(workers)

            self.assertEqual(len(worker_ranking), 1)

            best_worker, next_available, score, *_ = policies.mx.Last().apply(
                worker_ranking)

            self.assertEqual(best_worker.name, worker_01.name)
            self.assertEqual(best_worker.ip, worker_01.ip)
            self.assertEqual(score, 0.0)

    def test_pool_match_strip(self):
        worker_01 = Worker.objects.create(
            name='worker_01', ip='10.0.0.1',
            policies_settings={
                'pool': {'X_POOL_HEADER': 'X-MAILSEND-Pool', 'pools': [
                    'jambon']}})
        Worker.objects.create(
            name='worker_02', ip='10.0.0.2',
            policies_settings={
                'pool': {
                    'pools': ['foo'],
                    'X_POOL_HEADER': 'X-MAILSEND-Pool'}})
        mail_metadata_01 = Mail.objects.create(
            identifier='0001', headers={'X-MAILSEND-Pool': 'JAmbon  '})

        with fake_time('2015-12-10 12:00:00'):
            MailStatus.objects.create(
                destination_domain='example.com',
                mail=mail_metadata_01, source_ip=worker_01.ip)

        with fake_time('2015-12-10 12:00:05'):
            workers = policies.mx.First().apply({'To': 'test+01@example.com'})
            worker_ranking = pool.Policy(
                mail_metadata_01.identifier,
                {'To': 'test+01@example.com', 'X-MAILSEND-Pool': 'JAmbon  '},
                MailStatus).apply(workers)

            self.assertEqual(len(worker_ranking), 1)

            best_worker, next_available, score, *_ = policies.mx.Last().apply(
                worker_ranking)

            self.assertEqual(best_worker.name, worker_01.name)
            self.assertEqual(best_worker.ip, worker_01.ip)
            self.assertEqual(score, 0.0)



import sys

from django.core.management.base import BaseCommand

from ...utils import dkim


class Command(BaseCommand):
    help = 'DKIM sign .eml'

    def add_arguments(self, parser):
        parser.add_argument('selector', type=str)
        parser.add_argument('domain', type=str)
        parser.add_argument('privatekeyfile', type=str)
        parser.add_argument('-i', '--identity', type=str, required=False)

    def handle(self, *args, **options):
        sys.stdin = sys.stdin.detach()
        sys.stdout = sys.stdout.detach()

        identity = options['identity']

        domain = options['domain'].encode('utf-8')
        selector = options['selector'].encode('utf-8')
        privatekeyfile = options['privatekeyfile']

        if identity:
            identity = identity.encode('utf-8')

        message = sys.stdin.read()
        try:
            HEADERS_TO_SIGN = [
                'From', 'Subject', 'Reply-To', 'To', 'Date',
                'Message-ID', 'List-ID', 'List-Unsubscribe',
                'Sender', 'Content-Type', 'MIME-Version']
            _ENCODED_HEADERS_TO_SIGN = [
                h.encode('utf-8') for h in HEADERS_TO_SIGN]
            sig = dkim.sign(
                message, selector, domain,
                open(privatekeyfile, "rb").read(),
                identity=identity, include_headers=_ENCODED_HEADERS_TO_SIGN)
            sys.stdout.write(sig + b'\n')
            sys.stdout.write(message)
        except Exception as e:
            raise
            print(e, file=sys.stderr)
            sys.stdout.write(message)



import copy

from django.conf import settings
from django.utils import timezone
from django.test import override_settings
from libfaketime import fake_time

from munch_mailsend import policies
from munch_mailsend.models import Mail
from munch_mailsend.models import Worker
from munch_mailsend.models import MailStatus
from munch_mailsend.policies.mx import warm_up

from . import MailSendTestCase


class WarmUpPolicyTestCase(MailSendTestCase):
    def setUp(self):
        super().setUp()
        self.default_settings = {
            'warm_up': {
                'prioritize': 'equal',
                'domain_warm_up': {
                    'matrix': [5, 10, 30, 50, 100],
                    'goal': 50, 'step_tolerance': 10, 'max_tolerance': 10},
                'ip_warm_up': {
                    'enabled': True,
                    'matrix': [5, 10, 30, 50, 100],
                    'goal': 50, 'step_tolerance': 10, 'max_tolerance': 10}
            }
        }

    def create_mails(
            self, worker, number, destination_domain=None, warm_up_domains={}):
        now = timezone.now()
        MAILSEND_SETTINGS = copy.deepcopy(settings.MAILSEND)
        MAILSEND_SETTINGS['WORKER_POLICIES'] = [
            'munch_mailsend.policies.mx.warm_up.Policy']
        if warm_up_domains:
            MAILSEND_SETTINGS['WARM_UP_DOMAINS'] = warm_up_domains
        with override_settings(MAILSEND=MAILSEND_SETTINGS):
            for i in range(0, number):
                identifier = "{}.{}.{}".format(worker.name, i, now.timestamp())
                mail = Mail.objects.create(
                    headers={
                        'To': 'you@{}'.format(
                            destination_domain or 'example.com'),
                        'X-MAILSEND-Message-Id': identifier},
                    identifier=identifier)
                MailStatus.objects.create(
                    destination_domain=destination_domain or 'example.com',
                    mail=mail, source_ip=worker.ip,
                    status=MailStatus.QUEUED)

                best_worker, next_available, score, *_ = Worker.objects.\
                    find_worker(identifier, mail.headers, MailStatus)

                MailStatus.objects.create(
                    destination_domain=destination_domain or 'example.com',
                    mail=mail,
                    status=MailStatus.SENDING,
                    source_ip=worker.ip)

                MailStatus.objects.create(
                    destination_domain=destination_domain or 'example.com',
                    mail=mail,
                    status=MailStatus.DELIVERED,
                    source_ip=worker.ip)

    def test_simple_priority(self):
        """ Simple priority between two workers """
        worker_01 = Worker.objects.create(
            name='worker_01', ip='10.0.0.1',
            policies_settings=self.default_settings)
        worker_02 = Worker.objects.create(
            name='worker_02', ip='10.0.0.2',
            policies_settings=self.default_settings)

        with fake_time('2015-12-09 12:00:00'):
            self.create_mails(worker_01, 5)
            self.create_mails(worker_02, 5)

        with fake_time('2015-12-10 12:00:00'):
            self.create_mails(worker_01, 10)
            self.create_mails(worker_02, 9)

        with fake_time('2015-12-11 12:00:00'):
            self.create_mails(worker_01, 2)

        with fake_time('2015-12-11 18:30:30'):
            mail_metadata = Mail.objects.create(identifier="test-du-jambon")
            workers = policies.mx.First().apply({'To': 'test@example.com'})
            worker_ranking = warm_up.Policy(
                mail_metadata.identifier,
                {'To': 'test@example.com'}, MailStatus).apply(workers)

            best_worker, next_available, *_ = policies.mx.Last().apply(
                worker_ranking)
            self.assertEqual(best_worker.name, worker_02.name)
            self.assertEqual(best_worker.ip, worker_02.ip)

    def test_reached_steps_are_discared(self):
        """ No yesterday stats, then workers must start at first step """
        worker_01 = Worker.objects.create(
            name='worker_01', ip='10.0.0.1',
            policies_settings=self.default_settings)
        worker_02 = Worker.objects.create(
            name='worker_02', ip='10.0.0.2',
            policies_settings=self.default_settings)

        with fake_time('2015-12-09 12:00:00'):
            self.create_mails(worker_01, 5)
            self.create_mails(worker_02, 5)

        with fake_time('2015-12-10 12:00:00'):
            self.create_mails(worker_01, 11)
            self.create_mails(worker_02, 11)

        with fake_time('2015-12-10 18:30:30'):
            mail_metadata = Mail.objects.create(identifier="test_one")
            workers = policies.mx.First().apply({'To': 'test@example.com'})
            worker_ranking = warm_up.Policy(
                mail_metadata.identifier,
                {'To': 'test@example.com'}, MailStatus).apply(workers)

            self.assertEqual(len(worker_ranking), 0)
            best_worker, next_available, *_ = policies.mx.Last().apply(
                worker_ranking)
            self.assertIsNone(best_worker)

    def test_next_step(self):
        """ Reach each step every day (testing step_tolerance by the way) """
        worker = Worker.objects.create(
            name='worker', ip='10.0.0.1',
            policies_settings=self.default_settings)

        with fake_time('2015-12-10 12:00:00'):
            self.create_mails(worker, 4)

        with fake_time('2015-12-10 12:00:30'):
            mail_metadata = Mail.objects.create(identifier="1")
            workers = policies.mx.First().apply({'To': 'test@example.com'})
            worker_ranking = warm_up.Policy(
                mail_metadata.identifier,
                {'To': 'test@example.com'}, MailStatus).apply(workers)
            best_worker, next_available, *_ = policies.mx.Last().apply(
                worker_ranking)
            self.assertEqual(best_worker.name, worker.name)
            self.assertEqual(best_worker.ip, worker.ip)

        with fake_time('2015-12-11 12:00:00'):
            self.create_mails(worker, 10)

        with fake_time('2015-12-11 12:00:30'):
            mail_metadata = Mail.objects.create(identifier="2")
            workers = policies.mx.First().apply({'To': 'test@example.com'})
            worker_ranking = warm_up.Policy(
                mail_metadata.identifier,
                {'To': 'test@example.com'}, MailStatus).apply(workers)
            best_worker, next_available, *_ = policies.mx.Last().apply(
                worker_ranking)
            self.assertEqual(best_worker.name, worker.name)
            self.assertEqual(best_worker.ip, worker.ip)

        with fake_time('2015-12-12 12:00:00'):
            self.create_mails(worker, 32)
        with fake_time('2015-12-12 12:00:30'):
            mail_metadata = Mail.objects.create(identifier="3")
            workers = policies.mx.First().apply({'To': 'test@example.com'})
            worker_ranking = warm_up.Policy(
                mail_metadata.identifier,
                {'To': 'test@example.com'}, MailStatus).apply(workers)
            best_worker, next_available, *_ = policies.mx.Last().apply(
                worker_ranking)
            self.assertEqual(best_worker.name, worker.name)
            self.assertEqual(best_worker.ip, worker.ip)

        with fake_time('2015-12-13 12:00:00'):
            self.create_mails(worker, 49)
        with fake_time('2015-12-13 12:00:30'):
            mail_metadata = Mail.objects.create(identifier="4")
            workers = policies.mx.First().apply({'To': 'test@example.com'})
            worker_ranking = warm_up.Policy(
                mail_metadata.identifier,
                {'To': 'test@example.com'}, MailStatus).apply(workers)
            best_worker, next_available, *_ = policies.mx.Last().apply(
                worker_ranking)
            self.assertEqual(best_worker.name, worker.name)
            self.assertEqual(best_worker.ip, worker.ip)

        with fake_time('2015-12-14 12:00:00'):
            self.create_mails(worker, 54)
        with fake_time('2015-12-14 12:00:30'):
            mail_metadata = Mail.objects.create(identifier="5")
            workers = policies.mx.First().apply({'To': 'test@example.com'})
            worker_ranking = warm_up.Policy(
                mail_metadata.identifier,
                {'To': 'test@example.com'}, MailStatus).apply(workers)
            best_worker, next_available, *_ = policies.mx.Last().apply(
                worker_ranking)
            self.assertEqual(best_worker.name, worker.name)
            self.assertEqual(best_worker.ip, worker.ip)

        with fake_time('2015-12-14 12:00:35'):
            self.create_mails(worker, 1)
            mail_metadata = Mail.objects.create(identifier="6")
            workers = policies.mx.First().apply({'To': 'test@example.com'})
            worker_ranking = warm_up.Policy(
                mail_metadata.identifier,
                {'To': 'test@example.com'}, MailStatus).apply(workers)
            best_worker, next_available, *_ = policies.mx.Last().apply(
                worker_ranking)
            self.assertIsNone(best_worker)

    def test_max_step_tolerance(self):
        worker = Worker.objects.create(
            name='worker_01', ip='10.0.0.1',
            policies_settings={
                'warm_up': {
                    'ip_warm_up': {
                        'matrix': [5, 10, 30, 50, 100], 'goal': 50,
                        'step_tolerance': 10, 'max_tolerance': 0,
                        'enabled': True}}})

        with fake_time('2015-12-09 12:00:00'):
            self.create_mails(worker, 5)

        with fake_time('2015-12-10 12:00:00'):
            self.create_mails(worker, 10)

        with fake_time('2015-12-11 12:00:00'):
            self.create_mails(worker, 30)

        with fake_time('2015-12-11 18:30:30'):
            mail_metadata = Mail.objects.create(identifier="test_one")
            workers = policies.mx.First().apply({'To': 'test@example.com'})
            worker_ranking = warm_up.Policy(
                mail_metadata.identifier, {'To': 'test@example.com'},
                MailStatus).apply(workers)

        best_worker, next_available, *_ = policies.mx.Last().apply(
            worker_ranking)
        self.assertIsNone(best_worker)
        self.assertIsNone(next_available)

    def test_prioritize_warmest(self):
        worker_01 = Worker.objects.create(
            name='worker_01', ip='10.0.0.1', policies_settings={
                'warm_up': {
                    'prioritize': 'warmest',
                    'ip_warm_up': {
                        'matrix': [5, 10, 30, 50, 100], 'goal': 50,
                        'step_tolerance': 10,
                        'max_tolerance': 10, 'enabled': True}}})
        worker_02 = Worker.objects.create(
            name='worker_02', ip='10.0.0.2', policies_settings={
                'warm_up': {
                    'prioritize': 'warmest',
                    'ip_warm_up': {
                        'matrix': [5, 10, 30, 50, 100], 'goal': 50,
                        'step_tolerance': 10,
                        'max_tolerance': 10, 'enabled': True}}})

        with fake_time('2015-12-10 12:00:00'):
            self.create_mails(worker_01, 5)
            self.create_mails(worker_02, 5)

        with fake_time('2015-12-11 12:00:00'):
            self.create_mails(worker_01, 5)
            self.create_mails(worker_02, 3)

        with fake_time('2015-12-11 12:05:00'):
            MAILSEND_SETTINGS = copy.deepcopy(settings.MAILSEND)
            MAILSEND_SETTINGS['WORKER_POLICIES_SETTINGS'][
                'warm_up']['prioritize'] = 'warmest'
            with override_settings(MAILSEND=MAILSEND_SETTINGS):
                mail = Mail.objects.create(
                    identifier="1",
                    headers={
                        'To': 'test@example.com',
                        'X-MAILSEND-Message-Id': '1'})
                workers = policies.mx.First().apply(mail.headers)
                worker_ranking = warm_up.Policy(
                    mail.identifier, mail.headers,
                    MailStatus).apply(workers)
                best_worker, next_available, *_ = policies.mx.Last().apply(
                    worker_ranking)
                self.assertEqual(best_worker.name, worker_02.name)
                self.assertEqual(best_worker.ip, worker_02.ip)

    def test_prioritize_coldest(self):
        worker_01 = Worker.objects.create(
            name='worker_01', ip='10.0.0.1', policies_settings={
                'warm_up': {
                    'prioritize': 'coldest',
                    'ip_warm_up': {
                        'matrix': [5, 10, 30, 50, 100], 'goal': 50,
                        'step_tolerance': 10,
                        'max_tolerance': 10, 'enabled': True}}})
        worker_02 = Worker.objects.create(
            name='worker_02', ip='10.0.0.2', policies_settings={
                'warm_up': {
                    'prioritize': 'coldest',
                    'ip_warm_up': {
                        'matrix': [5, 10, 30, 50, 100], 'goal': 50,
                        'step_tolerance': 10,
                        'max_tolerance': 10, 'enabled': True}}})

        with fake_time('2015-12-10 12:00:00'):
            self.create_mails(worker_01, 5)
            self.create_mails(worker_02, 5)

        with fake_time('2015-12-11 12:00:00'):
            self.create_mails(worker_01, 1)
            self.create_mails(worker_02, 8)

        with fake_time('2015-12-11 12:05:00'):
            MAILSEND_SETTINGS = copy.deepcopy(settings.MAILSEND)
            MAILSEND_SETTINGS['WORKER_POLICIES_SETTINGS'][
                'warm_up']['prioritize'] = 'coldest'
            with override_settings(MAILSEND=MAILSEND_SETTINGS):
                mail = Mail.objects.create(
                    identifier="1",
                    headers={
                        'To': 'test@example.com',
                        'X-MAILSEND-Message-Id': '1'})
                workers = policies.mx.First().apply(mail.headers)
                worker_ranking = warm_up.Policy(
                    mail.identifier, mail.headers,
                    MailStatus).apply(workers)
                best_worker, next_available, *_ = policies.mx.Last().apply(
                    worker_ranking)
                self.assertEqual(best_worker.name, worker_01.name)
                self.assertEqual(best_worker.ip, worker_01.ip)

    def test_by_source_ip_disabled(self):
        worker = Worker.objects.create(
            name='worker_01', ip='10.0.0.1', policies_settings={
                'warm_up': {
                    'domain_warm_up': {
                        'matrix': [5, 10, 30, 50, 100], 'goal': 50,
                        'step_tolerance': 0, 'max_tolerance': 0},
                    'ip_warm_up': {
                        'matrix': [5, 10, 30, 50, 100], 'goal': 50,
                        'step_tolerance': 0, 'max_tolerance': 0,
                        'enabled': False}}})

        with fake_time('2015-12-09 12:00:00'):
            self.create_mails(worker, 5)

        with fake_time('2015-12-10 12:00:30'):
            mail = Mail.objects.create(identifier="1")
            workers = policies.mx.First().apply({'To': 'test@example.com'})
            worker_ranking = warm_up.Policy(
                mail.identifier, {'To': 'test@example.com'},
                MailStatus).apply(workers)
            best_worker, next_available, *_ = policies.mx.Last().apply(
                worker_ranking)
            self.assertEqual(best_worker.name, worker.name)
            self.assertEqual(best_worker.ip, worker.ip)

        with fake_time('2015-12-10 12:05:00'):
            self.create_mails(worker, 10, 'example.com', {
                'example.com': ('example.com')})
            MAILSEND_SETTINGS = copy.deepcopy(settings.MAILSEND)
            MAILSEND_SETTINGS['WORKER_POLICIES'] = [
                'munch_mailsend.policies.mx.warm_up.Apply']
            MAILSEND_SETTINGS['WARM_UP_DOMAINS'] = {
                'example.com': ('example.com')}
            with override_settings(MAILSEND=MAILSEND_SETTINGS):
                mail = Mail.objects.create(identifier="2")
                workers = policies.mx.First().apply({'To': 'test@example.com'})
                worker_ranking = warm_up.Policy(
                    mail.identifier, {'To': 'test@example.com'},
                    MailStatus).apply(workers)
                best_worker, next_available, *_ = policies.mx.Last().apply(
                    worker_ranking)
                self.assertIsNone(best_worker)

                mail = Mail.objects.create(identifier="3")
                workers = policies.mx.First().apply(
                    {'To': 'test@example2.com'})
                worker_ranking = warm_up.Policy(
                    mail.identifier, {'To': 'test@example2.com'},
                    MailStatus).apply(workers)
                best_worker, next_available, *_ = policies.mx.Last().apply(
                    worker_ranking)
                self.assertEqual(best_worker.name, worker.name)
                self.assertEqual(best_worker.ip, worker.ip)

    def test_by_domains_list(self):
        worker = Worker.objects.create(
            name='worker_01', ip='10.0.0.1',
            policies_settings={
                'warm_up': {
                    'domain_warm_up': {
                        'matrix': [5, 10, 30, 50, 100], 'goal': 50,
                        'step_tolerance': 0, 'max_tolerance': 0},
                    'ip_warm_up': {
                        'matrix': [5, 10, 30, 50, 100], 'goal': 50,
                        'step_tolerance': 0, 'max_tolerance': 0,
                        'enabled': False}
                }
            })

        with fake_time('2015-12-10 12:00:00'):
            MAILSEND_SETTINGS = copy.deepcopy(settings.MAILSEND)
            MAILSEND_SETTINGS['WARM_UP_DOMAINS'] = {
                'jambon': ('jambon.com', 'jambon.fr'),
                'gmail': ('gmail.com', 'gmail.fr')}
            with override_settings(MAILSEND=MAILSEND_SETTINGS):
                self.create_mails(worker, 5, destination_domain="gmail.com")

        with fake_time('2015-12-10 12:00:30'):
            MAILSEND_SETTINGS = copy.deepcopy(settings.MAILSEND)
            MAILSEND_SETTINGS['WARM_UP_DOMAINS'] = {
                'jambon': ('jambon.com', 'jambon.fr'),
                'gmail': ('gmail.com', 'gmail.fr')}
            with override_settings(MAILSEND=MAILSEND_SETTINGS):
                mail_metadata = Mail.objects.create(identifier="1")
                workers = policies.mx.First().apply({'To': 'test@gmail.fr'})
                worker_ranking = warm_up.Policy(
                    mail_metadata.identifier, {'To': 'test@gmail.fr'},
                    MailStatus).apply(workers)
                best_worker, *_ = policies.mx.Last().apply(
                    worker_ranking)
                self.assertIsNone(best_worker)

        with fake_time('2015-12-10 12:00:30'):
            mail_metadata = Mail.objects.create(identifier="2")
            workers = policies.mx.First().apply({'To': 'test@jambon.com'})
            worker_ranking = warm_up.Policy(
                mail_metadata.identifier, {'To': 'test@jambon.com'},
                MailStatus).apply(workers)
            best_worker, next_available, *_ = policies.mx.Last().apply(
                worker_ranking)
            self.assertEqual(best_worker.name, worker.name)



import logging

from celery import current_app as app
from django.conf import settings
from django.core.management.base import BaseCommand
from amqp.exceptions import NotFound

from ...models import Worker
from ...amqp import get_queue
from ...amqp import get_queue_size

log = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'List all queues'

    def print_queue_line(self, name, details=''):
        try:
            queue = get_queue(self.connection, name)
            self.stdout.write(self.line_format.format(
                queue.name, get_queue_size(queue), details))
        except NotFound:
            self.print_not_found_queue(name)

    def print_worker_queue_line(self, worker, details='', retry=False):
        try:
            count = 0
            queue = worker.get_queue(self.connection, retry=retry)
            if self.scheduled_tasks:
                if self.scheduled_tasks:
                    for t in self.scheduled_tasks.get(worker.name):
                        if t.get('request', {}).get(
                                'delivery_info', {}).get(
                                'routing_key') == queue.name:
                            count += 1
            else:
                details += '(worker too busy to request scheduled tasks)'
            if count:
                details += '({} scheduled)'.format(count)
            self.stdout.write(self.line_format.format(
                queue.name, worker.get_queue_size(queue), details))
        except NotFound:
            self.print_not_found_queue(queue.name)

    def print_not_found_queue(self, name):
        self.stdout.write(self.line_format.format(
            name, 'n/a', self.style.ERROR('(not created)')))

    def handle(self, *args, **options):
        self.line_format = '{:35}: {} {}'
        self.connection = app.connection()
        self.print_queue_line(settings.CELERY_DEFAULT_QUEUE)
        self.print_queue_line(settings.MAILSEND.get('ROUTING_QUEUE'))
        self.print_queue_line(settings.MAILSEND.get(
            'QUEUED_MAIL_QUEUE'))

        self.stdout.write('')
        for worker in Worker.objects.all():
            self.scheduled_tasks = {}
            scheduled_tasks_count = 0
            while not any(
                    self.scheduled_tasks.values()) \
                    and scheduled_tasks_count <= 2:
                scheduled_tasks_count += 1
                self.scheduled_tasks = app.control.inspect(
                    [worker.name]).scheduled() or {}
            details = self.style.ERROR(
                '(disabled) ') if not worker.enabled else ''
            self.print_worker_queue_line(worker, details)
            self.print_worker_queue_line(worker, details, retry=True)



from django.conf import settings
from kombu import Queue
from kombu import Exchange


def get_queue(connection, name):
    exchange = Exchange(
        channel=connection,
        name=settings.CELERY_DEFAULT_EXCHANGE,
        type=settings.CELERY_DEFAULT_EXCHANGE_TYPE)
    return Queue(name, exchange=exchange).bind(connection)


def get_queue_size(queue):
    return queue.queue_declare(passive=True).message_count



import math


class ExponentialBackOff:
    """ Configurable exponential Backoff function

    See https://slimta.org/latest/api/slimta.queue.html
    This one do not care about the envelope, only relies on timings

    Basic formula is A*e**n
    """
    _base = 250

    def __init__(self, retry_policy):
        """
        :param retry_policy: inspired by postfix retry settings, a 3-keys dict:
          - min_retry_interval: Minimum time (secs) between two retries
          - max_retry_interval: Maximum time (secs) between two retries
          - time_before_drop: Time (secs) before we drop the mail
            and notify the sender
        """
        try:
            self.min_retry_interval = retry_policy['min_retry_interval']
            self.max_retry_interval = retry_policy['max_retry_interval']
            self.time_before_drop = retry_policy['time_before_drop']
        except KeyError:
            raise KeyError(
                'RETRY_POLICY must define ''"min_retry_interval", '
                '"max_retry_interval" and "time_before_drop')
        # for cases where the min_retry_interval is under self._base
        # ensure the difference is always at least 1
        self.base = min(self._base, self.min_retry_interval - 1)
        self.A = self.get_A()

    def get_A(self):
        return (self.min_retry_interval - self.base) / math.e

    def delay(self, attempts):
        return min(
            self.A * math.e**attempts + self.base,
            self.max_retry_interval)

    def __call__(self, attempts):
        total_delay = sum(self.delay(i) for i in range(attempts))
        if total_delay <= self.time_before_drop:
            return int(self.delay(attempts))
        else:
            return None



from django.db import models
from django.contrib.postgres.fields import JSONField
from django.contrib.postgres.fields import HStoreField
from django.utils.translation import ugettext_lazy as _
from slimta.envelope import Envelope
from kombu import Queue
from kombu import Exchange

from munch.core.mail.utils import mk_base64_uuid
from munch.core.mail.models import RawMail
from munch.core.mail.models import AbstractMail
from munch.core.mail.models import AbstractMailStatus

from .settings import settings
from .managers import WorkerManager


class Worker(models.Model):
    name = models.CharField(
        max_length=100,
        help_text="Celery worker name (nothing related with SMTP EHLO)")
    ip = models.GenericIPAddressField(unique=True)
    creation_date = models.DateTimeField(auto_now_add=True)
    update_date = models.DateTimeField(auto_now=True)
    enabled = models.BooleanField(default=True)
    policies_settings = JSONField(null=True, blank=True)

    objects = WorkerManager()

    def __str__(self):
        return '{} ({})'.format(self.name, self.ip)

    def get_queue_name(self, retry=False):
        base = settings.MAILSEND.get('MX_WORKER_QUEUE_PREFIX')
        if retry:
            base = settings.MAILSEND.get('MX_WORKER_QUEUE_RETRY_PREFIX')
        return base.format(ip=self.ip)

    def get_queue(self, connection, retry=False):
        exchange = Exchange(
            channel=connection,
            name=settings.CELERY_DEFAULT_EXCHANGE,
            type=settings.CELERY_DEFAULT_EXCHANGE_TYPE)
        return Queue(self.get_queue_name(
            retry=retry), exchange=exchange).bind(connection)

    def get_queue_size(self, queue):
        return queue.queue_declare(passive=True).message_count

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if self.enabled:
            WorkerManager().set_to_cache(self)
        else:
            WorkerManager().remove_from_cache(self)


def get_mail_identifier():
    return mk_base64_uuid('i-')


class Mail(AbstractMail):
    identifier = models.CharField(
        max_length=35, db_index=True, unique=True,
        default=get_mail_identifier, verbose_name=_('identifier'))
    headers = HStoreField(default={})
    sender = models.EmailField()
    message = models.ForeignKey(RawMail, on_delete=models.SET_NULL, null=True)

    @classmethod
    def get_envelope(cls, identifier):
        from .utils.tasks import get_envelope
        return get_envelope(identifier)

    def as_envelope(self):
        envelope = Envelope()
        headers = ""
        for key, value in self.headers.items():
            headers += "{}: {}\n".format(key, value)
        if self.message is None:
            raise Exception(
                "Can't build this envelope because "
                "there is no RawMail attached to it.")
        message = self.message.content or ""
        envelope.parse(headers.encode('utf-8') + message.encode('utf-8'))
        envelope.sender = self.sender
        envelope.recipients.append(self.recipient)
        return envelope


class MailStatus(AbstractMailStatus):
    mail = models.ForeignKey(Mail)

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if self.status in self.FINAL_STATES:
            self.mail.message = None
            self.mail.save()



import unittest

from munch_mailsend.models import Mail
from munch_mailsend.utils.dkim import verify
from munch_mailsend.policies.relay import dkim

from . import MailSendTestCase

DNS_TXT = (
    'v=DKIM1;t=s;p=MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDt7Nt5JM5cu/uh8izxLyT'
    'xYf5HhgGDOE20S4XxpgGOX14DetiDhTX4ziyELgip6sC4vOQWVx+uVkz2uuwd8QgY8N3MhJDc'
    'eyPpgXi73Er91d8WckY5Gh1exSkg5dCmz9GYzd1Ci1PXlX5mTN+90s5uYsLK1DLGhpsXFNdM5'
    'atGsQIDAQAB')


@unittest.skip('Must fix dkimpy to verify dkim signature')
class DKIMPolicyTestCase(MailSendTestCase):
    def test_dkim_verify(self):
        mail = Mail.objects.create(
            identifier='0001', message="My Body",
            headers={
                'To': 'test-to@example.com',
                'From': 'test-from@mailsend-test.com',
                'Subject': 'My Subject'})
        envelope = mail.as_envelope()
        signer = dkim.Sign()
        signer.apply(envelope)
        headers_data, message_data = envelope.flatten()
        self.assertTrue(verify(
            headers_data + b'\r\n' + message_data,
            dnsfunc=lambda name: DNS_TXT))



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-06-21 09:47
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('core', '0002_permissions'),
    ]

    operations = [
        migrations.CreateModel(
            name='AgainAnotherMail',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('message', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='test_mail', to='core.RawMail')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='AnotherMail',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('message', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='test_anothermail', to='core.RawMail')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import re
import time

from .exceptions import DuplicateTag
from .exceptions import InvalidTagSpec
from .exceptions import ValidationError
from .exceptions import MessageFormatError

# FWS  =  ([*WSP CRLF] 1*WSP) /  obs-FWS ; Folding white space  [RFC5322]
FWS = br'(?:(?:\s*\r?\n)?\s+)?'


def get_bit_size(x):
    """Return size of long in bits."""
    return len(bin(x)) - 2


def hash_headers(
        hasher, canonicalize_headers, headers, include_headers, sigheader):
    """ Update hash for signed message header fields. """
    sign_headers = select_headers(headers, include_headers)
    # The call to _remove() assumes that the signature b= only appears
    # once in the signature header
    cheaders = canonicalize_headers.canonicalize_headers(
        [(sigheader[0], sigheader[1])])
    # the dkim sig is hashed with no trailing crlf, even if the
    # canonicalization algorithm would add one.
    for x, y in sign_headers + [(x, y.rstrip()) for x, y in cheaders]:
        hasher.update(x)
        hasher.update(b':')
        hasher.update(y)
    return sign_headers


def select_headers(headers, include_headers):
    """Select message header fields to be signed/verified.

    >>> h = [('from','biz'),('foo','bar'),('from','baz'),('subject','boring')]
    >>> i = ['from','subject','to','from']
    >>> select_headers(h,i)
    [('from', 'baz'), ('subject', 'boring'), ('from', 'biz')]
    >>> h = [('From','biz'),('Foo','bar'),('Subject','Boring')]
    >>> i = ['from','subject','to','from']
    >>> select_headers(h,i)
    [('From', 'biz'), ('Subject', 'Boring')]
    """
    signature_headers = []
    last_index = {}
    for header in include_headers:
        assert header == header.lower()
        i = last_index.get(header, len(headers))
        while i > 0:
            i -= 1
            if header == headers[i][0].lower():
                signature_headers.append(headers[i])
                break
        last_index[header] = i
    return signature_headers


def validate_signature_fields(signature):
    """
    Validate DKIM-Signature fields.

    Basic checks for presence and correct formatting of mandatory fields.
    Raises a ValidationError if checks fail, otherwise returns None.

    @param sig: A dict mapping field keys to values.
    """
    mandatory_fields = (b'v', b'a', b'b', b'bh', b'd', b'h', b's')
    for field in mandatory_fields:
        if field not in signature:
            raise ValidationError("signature missing {}=".format(field))

    if signature[b'v'] != b'1':
        raise ValidationError("v= value is not 1 ({})".format(signature[b'v']))
    if re.match(br"[\s0-9A-Za-z+/]+=*$", signature[b'b']) is None:
        raise ValidationError(
            "b= value is not valid base64 ({})".format(signature[b'b']))
    if re.match(br"[\s0-9A-Za-z+/]+=*$", signature[b'bh']) is None:
        raise ValidationError(
            "bh= value is not valid base64 (%{})".format(signature[b'bh']))
    # Nasty hack to support both str and bytes... check for both the
    # character and integer values.
    if b'i' in signature and (
            not signature[b'i'].lower().endswith(signature[b'd'].lower()) or
            signature[b'i'][-len(
                signature[b'd'])-1] not in ('@', '.', 64, 46)):
        raise ValidationError(
            "i= domain is not a subdomain of d= (i={} d={})".format(
                signature[b'i'], signature[b'd']))
    if b'l' in signature and re.match(br"\d{,76}$", signature[b'l']) is None:
        raise ValidationError(
            "l= value is not a decimal integer ({})".format(signature[b'l']))
    if b'q' in signature and signature[b'q'] != b'dns/txt':
        raise ValidationError("q= value is not dns/txt ({})".format(
            signature[b'q']))
    now = int(time.time())
    slop = 36000		# 10H leeway for mailers with inaccurate clocks
    t_sign = 0
    if b't' in signature:
        if re.match(br"\d+$", signature[b't']) is None:
            raise ValidationError(
                "t= value is not a decimal integer ({})".format(
                    signature[b't']))
        t_sign = int(signature[b't'])
        if t_sign > now + slop:
            raise ValidationError("t= value is in the future ({})".format(
                signature[b't']))
    if b'x' in signature:
        if re.match(rb"\d+$", signature[b'x']) is None:
            raise ValidationError(
                "x= value is not a decimal integer ({})".format(
                    signature[b'x']))
        x_sign = int(signature[b'x'])
        if x_sign < now - slop:
            raise ValidationError(
                "x= value is past ({})".format(signature[b'x']))
        if x_sign < t_sign:
            raise ValidationError(
                "x= value is less than t= value (x={} t={})".format(
                    signature[b'x'], signature[b't']))


def rfc822_parse(message):
    """
    Parse a message in RFC822 format.

    @param message: The message in RFC822 format.
                    Either CRLF or LF is an accepted line separator.
    @return: Returns a tuple of (headers, body)
             where headers is a list of (name, value) pairs.
    The body is a CRLF-separated string.
    """
    headers = []
    lines = re.split(br"\r?\n", message)
    i = 0
    while i < len(lines):
        if len(lines[i]) == 0:
            # End of headers, return what we have plus the body,
            # excluding the blank line.
            i += 1
            break
        if lines[i][0] in ("\x09", "\x20", 0x09, 0x20):
            headers[-1][1] += lines[i] + b"\r\n"
        else:
            m = re.match(br"([\x21-\x7e]+?):", lines[i])
            if m is not None:
                headers.append([m.group(1), lines[i][m.end(0):] + b"\r\n"])
            elif lines[i].startswith(b"From "):
                pass
            else:
                raise MessageFormatError(
                    "Unexpected characters in RFC822 header: {}".format(
                        lines[i]))
        i += 1
    return (headers, b"\r\n".join(lines[i:]))


def parse_tag_value(tag_list):
    """
    Parse a DKIM Tag=Value list.

    Interprets the syntax specified by RFC4871 section 3.2.
    Assumes that folding whitespace is already unfolded.

    @param tag_list: A bytes string containing a DKIM Tag=Value list.
    """
    tags = {}
    tag_specs = tag_list.strip().split(b';')
    # Trailing semicolons are valid.
    if not tag_specs[-1]:
        tag_specs.pop()
    for tag_spec in tag_specs:
        try:
            key, value = tag_spec.split(b'=', 1)
        except ValueError:
            raise InvalidTagSpec(tag_spec)
        if key.strip() in tags:
            raise DuplicateTag(key.strip())
        tags[key.strip()] = value.strip()
    return tags



import logging
from time import sleep

from django.core.exceptions import ObjectDoesNotExist
from django_redis import get_redis_connection

log = logging.getLogger(__name__)

conn = get_redis_connection('default')


def acquire_lock(
        lock_name, blocking_timeout=5, timeout=60 * 5, interval=0.1):
    lock = None
    counter = 0
    while not lock and counter < blocking_timeout:
        lock = conn.set(lock_name, 'true', timeout)
        counter += interval
        sleep(interval)
    return lock


def release_lock(lock_name):
    return conn.delete(lock_name)


def record_status(mailstatus, identifier, ehlo=None, reply=None):
    from ..models import Mail

    try:
        mailstatus.mail
    except ObjectDoesNotExist:
        mailstatus.mail = Mail.objects.get(identifier=identifier)
        mailstatus.save()


def get_envelope(identifier):
    from ..models import Mail

    return Mail.objects.get(identifier=identifier).as_envelope()



import copy

from django.conf import settings
from django.utils import timezone
from django.test import override_settings
from libfaketime import fake_time
from django_redis import get_redis_connection

from munch_mailsend.models import Mail
from munch_mailsend.models import Worker
from munch_mailsend.policies.mx.warm_up import Policy as WarmUpPolicy
from munch_mailsend.policies.mx.rate_limit import Policy as RateLimitPolicy

from . import MailSendTestCase
from ..models import MailStatus

conn = get_redis_connection('default')


class MailStatusCacheTestCase(MailSendTestCase):
    def setUp(self):
        super().setUp()
        self.settings = {
            'warm_up': {
                'prioritize': 'equal',
                'domain_warm_up': {
                    'matrix': [5, 10, 30, 50, 100],
                    'goal': 50, 'step_tolerance': 10, 'max_tolerance': 10},
                'ip_warm_up': {
                    'enabled': True,
                    'matrix': [5, 10, 30, 50, 100],
                    'goal': 50, 'step_tolerance': 10, 'max_tolerance': 10}
            }
        }

    def create_mails(
            self, worker, number, destination_domain=None, warm_up_domains={}):
        now = timezone.now()
        MAILSEND_SETTINGS = copy.deepcopy(settings.MAILSEND)
        if warm_up_domains:
            MAILSEND_SETTINGS['WARM_UP_DOMAINS'] = warm_up_domains
        with override_settings(MAILSEND=MAILSEND_SETTINGS):
            for i in range(0, number):
                identifier = "{}.{}.{}".format(worker.name, i, now.timestamp())
                mail = Mail.objects.create(
                    headers={
                        'To': 'you@{}'.format(
                            destination_domain or 'example.com'),
                        'X-MAILSEND-Message-Id': identifier},
                    identifier=identifier,
                    recipient='you@{}'.format(destination_domain))
                MailStatus.objects.create(
                    destination_domain=destination_domain or 'example.com',
                    mail=mail, source_ip=worker.ip,
                    status=MailStatus.QUEUED)

                best_worker, next_available, score, *_ = Worker.objects.\
                    find_worker(identifier, mail.headers, MailStatus)

                MailStatus.objects.create(
                    destination_domain=destination_domain or 'example.com',
                    mail=mail,
                    status=MailStatus.SENDING,
                    source_ip=worker.ip)

                MailStatus.objects.create(
                    destination_domain=destination_domain or 'example.com',
                    mail=mail,
                    status=MailStatus.DELIVERED,
                    source_ip=worker.ip)

    def test_build_cache(self):
        worker = Worker.objects.create(
            name='worker_01', ip='10.0.0.1', policies_settings=self.settings)
        warm_up_domains = {'foo': ['bar.com', 'foo.com']}
        MAILSEND_SETTINGS = copy.deepcopy(settings.MAILSEND)
        MAILSEND_SETTINGS['WARM_UP_DOMAINS'] = warm_up_domains

        with fake_time('2015-12-10 12:00:00'):
            with override_settings(MAILSEND=MAILSEND_SETTINGS):
                now = timezone.now()
                self.create_mails(
                    worker, 5,
                    destination_domain='bar.com',
                    warm_up_domains=warm_up_domains)
                self.assertEqual(
                    len(RateLimitPolicy.get_statuses(
                        worker.ip, 'bar.com', now)), 5)
                self.assertEqual(
                    WarmUpPolicy.get_step(worker.ip, 'foo'), (5, 0))
                for key in conn.scan_iter('{}:{}:*'.format(
                        settings.MAILSEND['CACHE_PREFIX'],
                        settings.MAILSEND['MAILSTATUS_CACHE_PREFIX'])):
                    conn.delete(key)
                self.assertEqual(
                    WarmUpPolicy.get_step(worker.ip, 'foo'), (None, None))
                MailStatus.objects.re_run_signals(
                    settings.MAILSEND['MAILSTATUS_CACHE_TIMEOUT'])
                self.assertEqual(
                    len(RateLimitPolicy.get_statuses(
                        worker.ip, 'bar.com', now)), 5)
                # Remains start at 0 because this is WarmUpPolicy which set it
                self.assertEqual(
                    WarmUpPolicy.get_step(worker.ip, 'foo'), (None, -5))

        with fake_time('2015-12-10 12:00:05'):
            warm_up_domains = {'foo': ['bar.com', 'foo.com']}
            with override_settings(MAILSEND=MAILSEND_SETTINGS):
                self.create_mails(
                    worker, 1, destination_domain='bar.com',
                    warm_up_domains=warm_up_domains)
                self.assertEqual(
                    WarmUpPolicy.get_step(worker.ip, 'foo'), (5, -1))

        with fake_time('2015-12-11 12:00:00'):
            with override_settings(MAILSEND=MAILSEND_SETTINGS):
                self.create_mails(
                    worker, 1, destination_domain='bar.com',
                    warm_up_domains=warm_up_domains)
                self.assertEqual(
                    WarmUpPolicy.get_step(worker.ip, 'foo'), (10, 10))

        with fake_time('2015-12-11 12:00:05'):
            with override_settings(MAILSEND=MAILSEND_SETTINGS):
                self.create_mails(
                    worker, 11, destination_domain='bar.com',
                    warm_up_domains=warm_up_domains)
                self.assertEqual(
                    WarmUpPolicy.get_step(worker.ip, 'foo'), (10, -1))
                for key in conn.scan_iter('{}:{}:*'.format(
                        settings.MAILSEND['CACHE_PREFIX'],
                        settings.MAILSEND['MAILSTATUS_CACHE_PREFIX'])):
                    conn.delete(key)
                self.assertEqual(
                    WarmUpPolicy.get_step(worker.ip, 'foo'), (None, None))
                MailStatus.objects.re_run_signals(
                    settings.MAILSEND['MAILSTATUS_CACHE_TIMEOUT'])
                # Remains start at 0 because this is WarmUpPolicy which set it
                self.assertEqual(
                    WarmUpPolicy.get_step(worker.ip, 'foo'), (None, -12))

        with fake_time('2015-12-12 12:00:00'):
            warm_up_domains = {'foo': ['bar.com', 'foo.com']}
            with override_settings(MAILSEND=MAILSEND_SETTINGS):
                self.create_mails(
                    worker, 1, destination_domain='bar.com',
                    warm_up_domains=warm_up_domains)
                self.assertEqual(
                    WarmUpPolicy.get_step(worker.ip, 'foo'), (30, 32))



from django.conf import settings
from slimta.policy import RelayPolicy


class StripBlacklisted(RelayPolicy):
    def apply(self, envelope):
        for header in settings.MAILSEND.get('BLACKLISTED_HEADERS', []):
            if header in envelope.headers:
                del envelope.headers[header]



import logging

from celery import current_app as app
from django.core.management.base import BaseCommand

from ...models import Worker

log = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'List all worker messages'

    def add_arguments(self, parser):
        parser.add_argument('ip', nargs='+', type=str)
        parser.add_argument(
            '--limit', dest='limit', default=50, type=int,
            action='store', help="Message display limit (Default: 50)")

    def handle(self, *args, **options):
        class Counter:
            pass
        counter = Counter()
        counter.count = 0

        workers = Worker.objects.filter(ip__in=options['ip'])

        if not workers:
            print('Error: workers not found')

        self.scheduled_tasks = {}
        scheduled_tasks_count = 0
        while not any(
                self.scheduled_tasks.values()) and scheduled_tasks_count <= 2:
            scheduled_tasks_count += 1
            self.scheduled_tasks = app.control.inspect(
                [w.name for w in workers]).scheduled() or {}

        print('{:36};{:25};{:22};eta'.format(
            "task-id", "task-name", "message-id"))
        for worker in workers:
            if self.scheduled_tasks:
                scheduled_tasks = self.scheduled_tasks.get(worker.name)
                for t in scheduled_tasks:
                    r = t.get('request')
                    task_id = r.get('id')
                    task_name = r.get('name')
                    eta = t.get('eta')
                    identifier = eval(t.get('request').get('args'))[0]
                    print('{};{};{};{}'.format(
                        task_id, task_name, identifier, eta))



import logging

from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand

log = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'Cache MailStatus'

    def add_arguments(self, parser):
        parser.add_argument('app_label', type=str)
        parser.add_argument('model_name', type=str)

    def handle(self, *args, **options):
        mailstatus_class = apps.get_model(
            options.get('app_label'), options.get('model_name'))
        count = mailstatus_class.objects.re_run_signals(
            settings.MAILSEND['MAILSTATUS_CACHE_TIMEOUT'])
        self.stdout.write(
            '{} MailStatus object(s) since {} second(s) have been '
            'cached. Done.'.format(
                count, settings.MAILSEND['MAILSTATUS_CACHE_TIMEOUT']))



from django.conf import settings
from django.test import override_settings

from munch_mailsend.models import Mail
from munch_mailsend.models import RawMail
from munch_mailsend.policies.relay import headers

from . import MailSendTestCase


class BlacklistedHeadersTestCase(MailSendTestCase):
    @override_settings(MAILSEND={'BLACKLISTED_HEADERS': ['jambon']})
    def test_blacklisted_headers(self):
        raw_mail, _ = RawMail.objects.get_or_create(content='My Body')
        mail = Mail.objects.create(
            identifier='0001', message=raw_mail,
            headers={'To': 'foo@bar', 'jambon': 'My Jambonnery year'})
        envelope = mail.as_envelope()
        cleaner = headers.StripBlacklisted()
        cleaner.apply(envelope)
        for field in settings.MAILSEND['BLACKLISTED_HEADERS']:
            self.assertNotIn(field, envelope.headers)

    @override_settings(MAILSEND={'BLACKLISTED_HEADERS': ['jaMbon']})
    def test_blacklisted_headers_insensitive_case(self):
        raw_mail, _ = RawMail.objects.get_or_create(content='My Body')
        value = 'My Jambonnnery yea'
        mail = Mail.objects.create(
            identifier='0001', message=raw_mail,
            headers={'To': 'foo@bar', 'jambon': value})
        envelope = mail.as_envelope()
        cleaner = headers.StripBlacklisted()
        cleaner.apply(envelope)
        for field in settings.MAILSEND['BLACKLISTED_HEADERS']:
            self.assertNotIn(field, envelope.headers)
        self.assertNotIn('jambon', envelope.headers)



from ssl import SSLContext
from ssl import PROTOCOL_TLSv1_2  # TODO: Switch to PROTOCOL_TLS with Py3.5

from gevent.socket import create_connection
from django.conf import settings
from django.utils.module_loading import import_string
from slimta.relay.smtp.mx import MxSmtpRelay as MxSmtpRelayBase


class RelayStartupFatalError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return 'FATAL, abording startup: {}'.format(self.msg)


def _socket_creator(address):
    return create_connection(
        address,
        source_address=(settings.MAILSEND.get('SMTP_WORKER_SRC_ADDR'), 0))


class MxSmtpRelay(MxSmtpRelayBase):
    def __init__(self, *args, **kwargs):
        self.ehlo = settings.MAILSEND.get('SMTP_WORKER_EHLO_AS')
        kwargs.setdefault('ehlo_as', self.ehlo)
        kwargs.setdefault('socket_creator', _socket_creator)

        if settings.MAILSEND.get('TLS'):
            ssl_context = SSLContext(PROTOCOL_TLSv1_2)
            ssl_context.load_cert_chain(
                settings.MAILSEND.get('TLS').get('certfile'),
                keyfile=settings.MAILSEND.get('TLS').get('keyfile'))
            kwargs.setdefault('context', ssl_context)

        for k, v in settings.MAILSEND.get('RELAY_TIMEOUTS', {}).items():
            kwargs.setdefault(k, v)

        # Add a binary_encoder if provided to convert utf-8 emails to ascii
        # for servers not supporting 8BITMIME
        kwargs.setdefault('binary_encoder', settings.MAILSEND.get(
            'BINARY_ENCODER'))

        super().__init__(*args, **kwargs)

        # Override MX lookups for specific configured domains
        # This might be used in Dev mode to avoid sending emails
        for force_mx in settings.MAILSEND.get('SMTP_WORKER_FORCE_MX', []):
            self.force_mx(**force_mx)

        self._add_settings_defined_policies()

    def _add_settings_defined_policies(self):
        for path in settings.MAILSEND.get('RELAY_POLICIES', []):
            try:
                policy = import_string(path)
            except ImportError:
                raise RelayStartupFatalError(
                    '{} points to inexistant policy'.format(path))
            except TypeError:
                raise RelayStartupFatalError(
                    '{} is not a valid RelayPolicy'.format(path))
            self.add_policy(policy())

    def attempt(self, envelope, *args, **kwargs):
        # Oddly, removing and re-adding the subject header prevents
        # bad encoding which result in utf-8 characters being encoded
        # like: =?unknown-8bit?q?
        # The documentation at https://docs.python.org/3.4/library/email.message.html#email.message.Message.add_header  # noqa
        # mentions that a header with non-ASCII character will be automatically
        # encoded using a utf-8 Charset. Likely related...
        # This should probably be addressed in Slimta though.
        subject = envelope.headers['Subject']
        del envelope.headers['Subject']
        envelope.headers.add_header('Subject', subject)
        # This is to avoid "Bare LF" errors (eg: with free.fr)
        # Probably not the cleanest way to do it but it works
        # TODO: see if this could be done upstream in envelope generation
        # or if slimta needs a fix (it should not let bare \n go through...)
        envelope.message = envelope.message.decode('utf-8').replace(
            '\n', '\r\n').encode('utf-8')
        # Attempt delivery as usual
        return super().attempt(envelope, *args, **kwargs)



from django.db import models

from django.db.models.signals import pre_save
from django.db.models.signals import post_save

from munch.core.mail import backend

from munch_mailsend.models import RawMail
from munch_mailsend.models import MailStatus


class MailSendModel(models.Model):
    """
    Base for test models that sets app_label, so they play nicely.
    """
    class Meta:
        app_label = 'mailsend_tests'
        abstract = True


class AnotherMail(MailSendModel):
    message = models.ForeignKey(
        RawMail, on_delete=models.SET_NULL,
        null=True, related_name='test_anothermail')


class AgainAnotherMail(MailSendModel):
    message = models.ForeignKey(
        RawMail, on_delete=models.SET_NULL,
        null=True, related_name='test_mail')


pre_save.connect(backend.pre_save_mailstatus_signal, sender=MailStatus)
post_save.connect(backend.post_save_mailstatus_signal, sender=MailStatus)



import logging

from django.conf import settings
from django.core.management.base import BaseCommand
from django_redis import get_redis_connection

log = logging.getLogger(__name__)

conn = get_redis_connection('default')


class Command(BaseCommand):
    help = 'Clear MailStatus cache'

    def handle(self, *args, **options):
        count = 0
        for key in conn.scan_iter('{}:{}:*'.format(
                settings.MAILSEND['CACHE_PREFIX'],
                settings.MAILSEND['MAILSTATUS_CACHE_PREFIX'])):
            count += conn.delete(key)
        self.stdout.write(
            '{} key(s) deleted. Done.'.format(count))



import logging
from datetime import datetime

import pytz
from django.conf import settings
from django.utils import timezone

from munch.core.mail.utils import extract_domain

from ...models import Worker

CACHE_PREFIX = '{}:{}'.format(
    settings.MAILSEND['CACHE_PREFIX'],
    settings.MAILSEND['MAILSTATUS_CACHE_PREFIX'])
CACHE_TIMEOUT = settings.MAILSEND['MAILSTATUS_CACHE_TIMEOUT']


class First:
    def apply(self, headers, not_before=None):
        workers = list(Worker.objects.get_from_cache())
        if not workers:
            for worker in Worker.objects.filter(enabled=True):
                Worker.objects.set_to_cache(worker)
            workers = list(Worker.objects.get_from_cache())
        for worker in workers:
            worker['score'] = 0.0
            worker['next_available'] = not_before or timezone.now()
        return workers


class Last:
    def apply(self, workers):
        if workers:
            worker = max(workers, key=lambda worker: worker['score'])
            return (
                Worker.objects.get(ip=worker.get('ip')),
                worker.get('next_available'),
                worker.get('score'),
                workers)
        return (None, None, None, {})


class WorkerPolicyException(Exception):
    pass


class WorkerPolicyBase:
    def __init__(
            self, identifier, headers, mail_status_class, reply_code=None,
            reply_message=None, not_before=None):
        self.headers = headers
        self.identifier = identifier
        self.not_before = not_before
        self.reply_code = reply_code
        self.reply_message = reply_message
        self.mail_statuses = mail_status_class
        self.logger = logging.getLogger(__name__)

    def get_settings(self, worker):
        return worker.get('policies_settings', {}).get(
            self.__module__.split('.')[-1], {})

    def old(self):
        return datetime(year=1970, month=1, day=1).replace(tzinfo=pytz.utc)

    def now(self):
        return datetime.utcnow().replace(tzinfo=pytz.utc)

    def get_domain(self, address):
        return extract_domain(address)

    def apply(self, workers):
        """
            This method must only return workers with updated:
                `next_available` and `score`
        """
        raise NotImplementedError

    ###########
    # Signals #
    ###########
    @classmethod
    def mailstatus_pre_save(cls, instance, manager):
        pass

    @classmethod
    def mailstatus_post_save(cls, instance, manager):
        pass



import logging

from django.core.management.base import BaseCommand

from ...models import Worker

log = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'List all workers'

    def handle(self, *args, **options):
        for worker in Worker.objects.all():
            color = self.style.SUCCESS if worker.enabled else self.style.ERROR
            self.stdout.write(color(
                '* {} (pk:{})'.format(worker.ip, worker.pk, worker.enabled)))
            print('├────────────── Details ──────────────')
            print('├─ Enabled: {}'.format(worker.enabled))
            print('├─ Name: {}'.format(worker.name))
            print('├─ Creation date: {}'.format(worker.creation_date))
            print('├─ Update date: {}'.format(worker.update_date))
            print('└─────────────────────────────────────')



import sys

from django.core.management.base import BaseCommand

from ...utils import dkim


class Command(BaseCommand):
    help = 'DKIM verify envelope'

    def handle(self, *args, **options):
        sys.stdin = sys.stdin.detach()
        message = sys.stdin.read()

        d = dkim.DKIM(message)
        res = d.verify()
        if not res:
            print("signature verification failed")
            sys.exit(1)
        print("signature ok")



import uuid
import socket
import logging
from random import randint
from datetime import timedelta

from celery import task
from celery import current_app
from django.conf import settings
from django.db import connection
from django.utils import timezone
from django.utils.module_loading import import_string
from slimta.smtp.reply import Reply
from slimta.relay.smtp.mx import PermanentRelayError
from slimta.relay.smtp.mx import TransientRelayError
from django_redis import get_redis_connection

from munch.core.mail.utils import extract_domain
from munch.core.utils.tasks import task_autoretry
from munch.core.mail.models import AbstractMailStatus
from munch.core.mail.exceptions import SoftFailure
from munch.core.utils import get_worker_types

from .utils import save_timer
from .utils import ExponentialBackOff
from .utils.tasks import acquire_lock
from .utils.tasks import release_lock
from .models import Worker
from .amqp import get_queue
from .amqp import get_queue_size
from .relay import MxSmtpRelay

log = logging.getLogger(__name__)
conn = get_redis_connection()
worker_types = get_worker_types()


@task_autoretry(
    autoretry_on=(Exception, ),
    default_retry_delay=settings.MAILSEND['TASKS_SETTINGS'][
        'send_email']['default_retry_delay'],
    max_retries=settings.MAILSEND['TASKS_SETTINGS'][
        'send_email']['max_retries'],
    retry_message='Error while trying to send email. Retrying.')
@save_timer(name='mailsend.tasks.send_email')
def send_email(
        identifier, headers, attempts, mailstatus_class_path,
        record_status_task_path, build_envelope_task_path, token=None):
    # Retrieve MailStatus class and record_status task
    mailstatus_class = import_string(mailstatus_class_path)
    record_status_task = import_string(record_status_task_path)
    build_envelope_task = import_string(build_envelope_task_path)

    # Helper to create a new MailStatus
    def record_new_status(status, identifier, headers, reply, ehlo):
        log.debug('[{}] [worker:{}] Recording "{}" status...'.format(
            identifier, settings.MAILSEND['SMTP_WORKER_SRC_ADDR'], status))
        mailstatus_kwargs = {}

        raw_message = ''
        if reply.code:
            raw_message += reply.code + ' '
        if reply.raw_message:
            raw_message += reply.raw_message

        if raw_message:
            mailstatus_kwargs.update({'raw_msg': raw_message.strip()})

        if reply.enhanced_status_code:
            mailstatus_kwargs.update(
                {'status_code': reply.enhanced_status_code})

        mailstatus = mailstatus_class(
            status=status,
            destination_domain=extract_domain(headers.get('To')),
            **mailstatus_kwargs)
        # Status recording needs to happen locally and synchronously:
        #   - status needs to be available immediately for future routing
        #   - Creation of object derived from AbstractMailStatus
        #     will use the local worker IP address as source
        #     (although we could pass it)
        try:
            record_status_task(mailstatus, identifier, ehlo, reply)
        except SoftFailure as exc:
            log.info(
                'SoftFailure during "send_email" task ('
                'discarding this task): {}'.format(str(exc)), exc_info=True)
            return
        if status in AbstractMailStatus.FINAL_STATES:
            delete_envelope_token(identifier)

    # Helper to properly handle a transient failure
    def handle_transient_failure(
            identifier, headers, attempts, reply, ehlo):
        log.debug(
            '[{}] Handling transient failure ({} {})'
            ' (attempts:{}).'.format(
                identifier, reply.code, reply.message, attempts))
        backoff = ExponentialBackOff(settings.MAILSEND.get('RETRY_POLICY'))
        wait = backoff(attempts + 1)
        if wait:
            not_before = timezone.now() + timedelta(seconds=wait)
            record_new_status(
                AbstractMailStatus.DELAYED,
                identifier, headers, reply, ehlo)
            route_envelope.apply_async(
                (
                    identifier, headers,
                    attempts + 1,
                    mailstatus_class_path,
                    record_status_task_path,
                    build_envelope_task_path),
                {'not_before': not_before, 'reply': reply})
        else:
            reply.message += ' (Too many retries)'
            record_new_status(
                AbstractMailStatus.DROPPED, identifier, headers, reply, ehlo)

    if not any([t in worker_types for t in ['mx', 'all']]):
        countdown = 60 * 10
        log.error(
            '[{}] [worker:{}] Received "send_email" task but '
            'this is not an MX worker ({}) (re-routing in {} minutes)'.format(
                identifier,
                settings.MAILSEND['SMTP_WORKER_SRC_ADDR'],
                worker_type, countdown / 60))
        reply = Reply(
            '450',
            (
                '4.0.0 Unhandled delivery error: Re-trying to send '
                'envelope in {} minutes.').format(countdown / 60))
        ehlo = settings.MAILSEND['SMTP_WORKER_SRC_ADDR']
        record_new_status(
            AbstractMailStatus.DELAYED, identifier, headers, reply, ehlo)
        route_envelope.apply_async(
            (
                identifier, headers, attempts,
                mailstatus_class_path,
                record_status_task_path,
                build_envelope_task_path),
            {'not_before': None, 'reply': None},
            countdown=countdown)
        return

    # Envelope with final states must be discards
    latest_status = mailstatus_class.objects.filter(
        status__in=[AbstractMailStatus.DELETED] +
        list(AbstractMailStatus.FINAL_STATES), mail__identifier=identifier)
    if latest_status:
        log.debug(
            "[{}] Envelope ignored because it has already been "
            "{} at {}".format(
                identifier, latest_status[0].status,
                latest_status[0].creation_date))
        return

    # If envelope doesn't have token in cache, there is a serious problem
    current_token = get_envelope_token(identifier)
    if current_token is None:
        reply = Reply(
            '450',
            '4.0.0 Unhandled delivery error: No envelope token found in cache')
        handle_transient_failure(
            identifier, headers, attempts,
            reply, settings.MAILSEND.get('SMTP_WORKER_EHLO_AS'))
        log.error(
            "[{}] Error while trying to get envelope token. "
            "Envelope will be re-routed.".format(identifier), exc_info=True)
        return
    # If token mismatch, maybe this task is a duplicate (problem incoming)
    if token != current_token:
        log.info(
            "[{}] Discarding these send_email task "
            "because token doesn't match")
        return
    log.debug('[{}] Token is valid: {}'.format(identifier, token))

    if attempts:
        log.debug(
            '[{}] [worker:{}] Retrying to send (attempts:{}) '
            '(from:{}) (to:{})...'.format(
                identifier, settings.MAILSEND['SMTP_WORKER_SRC_ADDR'],
                attempts, headers['From'], headers['To']))
    else:
        log.debug(
            '[{}] [worker:{}] Sending envelope '
            '(from:{}) (to:{})...'.format(
                identifier, settings.MAILSEND['SMTP_WORKER_SRC_ADDR'],
                headers['From'], headers['To']))

    try:
        relay = MxSmtpRelay()
        # We call parent (slimta.relay.Relay) _attempt()
        # to run relay policies
        try:
            envelope = build_envelope_task(identifier)
        except SoftFailure as exc:
            log.info(
                'SoftFailure during "send_email" task ('
                'discarding this task): {}'.format(str(exc)), exc_info=True)
            return
        reply = relay._attempt(envelope, attempts)
        # We assume single-recipient
        if isinstance(reply, (list, tuple)):
            reply = reply[0]
        elif isinstance(reply, dict):
            reply = list(reply.values())[0]
    except TransientRelayError as exc:
        handle_transient_failure(
            identifier, headers, attempts, exc.reply, relay.ehlo)
    except PermanentRelayError as exc:
        log.debug(
            '[{}] [worker:{}] Handling PermanentRelayError'
            'with reply: {}'.format(
                identifier,
                settings.MAILSEND['SMTP_WORKER_SRC_ADDR'],
                exc.reply))
        record_new_status(
            AbstractMailStatus.BOUNCED,
            identifier, headers, exc.reply, relay.ehlo)
    except (Exception, BrokenPipeError, IOError, OSError) as exc:
        reply = Reply('450', '4.0.0 Unhandled delivery error: ' + str(exc))
        handle_transient_failure(
            identifier, headers, attempts, reply, relay.ehlo)
        log.error(
            "[{}] Error while trying to send email via Slimta. "
            "Envelope will be re-routed.".format(
                identifier), exc_info=True)
    else:
        record_new_status(
            AbstractMailStatus.DELIVERED,
            identifier, headers, reply, relay.ehlo)


@task_autoretry(
    autoretry_on=(Exception, ),
    default_retry_delay=settings.MAILSEND['TASKS_SETTINGS'][
        'route_envelope']['default_retry_delay'],
    max_retries=settings.MAILSEND['TASKS_SETTINGS'][
        'route_envelope']['max_retries'],
    retry_message='Error while trying to route envelope. Retrying.')
@save_timer(name='mailsend.tasks.route_envelope')
def route_envelope(
        identifier, headers, attempts, mailstatus_class_path,
        record_status_task_path, build_envelope_task_path,
        not_before=None, reply=None):
    """
        This envelope routing task take initiate attempt
        to free Slimta Edge from SMTP connection
    """
    record_performance = settings.STATSD_ENABLED

    lock = None
    mailstatus_class = import_string(mailstatus_class_path)
    latest_status = mailstatus_class.objects.filter(
        status__in=[AbstractMailStatus.DELETED] +
        list(AbstractMailStatus.FINAL_STATES), mail__identifier=identifier)
    if latest_status:
        log.debug(
            "[{}] Envelope ignored because it has already been "
            "{} at {}".format(
                identifier, latest_status[0].status,
                latest_status[0].creation_date))
        return
    # Ensure we close Django database connection because we don't
    # want to have opened connections while waiting for lock.
    connection.close()

    destination_domain = extract_domain(headers.get('To', ''))

    pool = headers.get(settings.MAILSEND['X_POOL_HEADER'], 'default')

    lock_name = '{}:lock:routing:{}:{}'.format(
        settings.MAILSEND['CACHE_PREFIX'], destination_domain, pool)
    lock_timeout = settings.MAILSEND['ROUTER_LOCK_TIMEOUT']
    lock_blocking_timeout = settings.MAILSEND['ROUTER_LOCK_WAITING']
    log.debug(
        '[{}] Waiting for lock "{}" (timeout:{} '
        'second(s) and blocking for {} second(s))...'.format(
            identifier, lock_name, lock_timeout, lock_blocking_timeout))

    if record_performance:
        from statsd.defaults.django import statsd
        lock_timer = statsd.timer('mailsend.tasks.lock_waiting')
        lock_timer.start()

    lock = acquire_lock(
        lock_name, timeout=lock_timeout,
        blocking_timeout=lock_blocking_timeout)

    if record_performance:
        lock_timer.stop()

    if lock:
        try:
            log.debug('[{}] Routing envelope (attempts={})...'.format(
                identifier, attempts))

            worker, next_available, score, others = Worker.objects.find_worker(
                identifier, headers, mailstatus_class,
                not_before=not_before, reply=reply)
            if worker:
                log.debug(
                    '[{}] Choosen worker is available at {} '
                    'with a {} score. Full workers ranking: {}'.format(
                        identifier, next_available.astimezone(), score, {
                            w.get('ip'): {
                                'score': w.get('score'),
                                'next_available': str(
                                    w.get('next_available').astimezone())}
                            for w in others}))
                mail_status_kwargs = {
                    'source_ip': worker.ip,
                    'status': AbstractMailStatus.SENDING,
                    'destination_domain': extract_domain(
                        headers.get('To'))}
                if not attempts:
                    routing_key = worker.get_queue_name()
                else:
                    routing_key = worker.get_queue_name(retry=True)

                attempt = send_email.s(
                    identifier, headers, attempts,
                    mailstatus_class_path,
                    record_status_task_path,
                    build_envelope_task_path,
                    token=set_envelope_token(identifier))
                now = timezone.now()
                if next_available:
                    countdown = max(0, (next_available - now).total_seconds())
                # And apply countdown to task if > 0
                if countdown:
                    mail_status_kwargs.update({
                        'creation_date': now + timedelta(seconds=countdown)})
                    attempt.set(countdown=countdown)

                log.info(
                    '[{}] Queued with "{}" routing key in {} seconds'.format(
                        identifier, routing_key, int(countdown)))
                mailstatus = mailstatus_class(**mail_status_kwargs)
                record_status_task = import_string(record_status_task_path)
                try:
                    record_status_task(mailstatus, identifier, attempts + 1)
                except SoftFailure as exc:
                    log.info(
                        'SoftFailure during "route_envelope" task ('
                        'discarding this task): {}'.format(
                            str(exc)), exc_info=True)
                    release_lock(lock_name)
                    return

                release_lock(lock_name)

                return attempt.apply_async(routing_key=routing_key).id
            else:
                log.debug(
                    '[{}] No worker available. Re-route envelope '
                    'in 5 minutes'.format(identifier))
                release_lock(lock_name)
                return route_envelope.apply_async(
                    (
                        identifier, headers, attempts,
                        mailstatus_class_path,
                        record_status_task_path,
                        build_envelope_task_path),
                    {'not_before': not_before, 'reply': reply},
                    countdown=60 * 5).id
        except Exception:
            release_lock(lock_name)
            raise
    else:
        log.debug(
            '[{}] Failed to acquire lock after waiting {} second(s). '
            'Re-route task in 1-6 seconds.'.format(
                identifier, lock_blocking_timeout))
        return route_envelope.apply_async(
            (identifier, headers, attempts),
            {
                'mailstatus_class_path': mailstatus_class_path,
                'record_status_task_path': record_status_task_path,
                'build_envelope_task_path': build_envelope_task_path,
                'not_before': not_before, 'reply': reply},
            countdown=randint(1, 6)).id

    if lock:
        release_lock(lock_name)


@task
def ping_workers():
    """ Ping enabled worker and disable them if they doesn't respond a pong """
    from .models import Worker

    workers = Worker.objects.filter(enabled=True).only('pk', 'name')
    pings = [list(node)[0] for node in current_app.control.ping(timeout=3)]
    max_failure = settings.MAILSEND['MX_WORKER_MAX_PING_FAILURES']

    for worker in workers:
        log.debug('[{}] Pinging worker... (timeout=3)'.format(worker.ip))
        key = '{}:worker:ping_failures:{}'.format(
            settings.MAILSEND.get('CACHE_PREFIX'), worker.ip)
        if worker.name not in pings:
            failure_counter = conn.get(key)
            if failure_counter is None:
                conn.set(key, 0, 60 * 5)
                failure_counter = 0
            log.debug('[{}] No response (failure={},max={})'.format(
                worker.ip, failure_counter, max_failure))
            if failure_counter > max_failure:
                log.warn(
                    "[{}] Worker (pk:{}) seems to have crashed because it "
                    "doesn't have updated its status. Disabling it...".format(
                        worker.name, worker.pk))
                worker.enabled = False
                worker.save()
                conn.delete(key)
            else:
                log.debug(
                    "[{}] This is a soft warning, let's see "
                    "next ping...".format(worker.ip))
                conn.incr(key)
        else:
            log.debug('[{}] Worker is up'.format(worker.ip))
            conn.delete(key)


@task
def check_disabled_workers():
    """
    Check if disabled workers have tasks in queue, then re-route them
    """
    from .models import Worker
    from .tasks import route_envelope

    class Counter:
        pass
    c = current_app.connection()
    counter = Counter()
    counter.count = 0

    def process_message(body, message):
        identifier = body.get('args', [])[0]
        attempts = body.get('args')[1]
        log.info(
            '[{}] Republishing mail into routing task (attempts={})...'.format(
                identifier, attempts))
        route_envelope.apply_async(body.get('args'))
        message.ack()
        counter.count += 1

    for worker in Worker.objects.filter(enabled=False).only('ip'):
        queues = [
            worker.get_queue(connection),
            worker.get_queue(connection, retry=True)]
        for queue in queues:
            queue = worker.get_queue(connection)
            size = worker.get_queue_size(queue)
            if size:
                log.info(
                    "{} tasks remaining in disabled queue {}. Republishing them...".format(
                        size, queue.name))
                task_consumer = current_app.amqp.TaskConsumer(
                    c, queues=[queue], callbacks=[process_message])
                with task_consumer:
                    while counter.count <= size:
                        try:
                            connection.drain_events(timeout=2)
                        except socket.timeout:
                            return


@task
def dispatch_queued():
    """
    Reroute all queued tasks
    This task is not used for now but maybe a management command
    could be usefull.
    """
    from .tasks import route_envelope

    class Counter:
        pass
    connection = current_app.connection()
    counter = Counter()
    counter.count = 0

    def process_message(body, message):
        identifier = body.get('args', [])[0]
        attempts = body.get('args')[1]
        log.info(
            'Republishing {} mail into routing task (attempts={})...'.format(
                identifier, attempts))
        message.ack()
        route_envelope.apply_async(body.get('args'))
        counter.count += 1

    queue = get_queue(connection, settings.MAILSEND['QUEUED_MAIL_QUEUE'])
    size = get_queue_size(queue)
    if size:
        log.info("Rerouting {} tasks from {} queue...".format(
            size, settings.MAILSEND['QUEUED_MAIL_QUEUE']))
        task_consumer = current_app.amqp.TaskConsumer(
            connection, queues=[queue], callbacks=[process_message])
        with task_consumer:
            while counter.count <= size:
                try:
                    connection.drain_events(timeout=2)
                except socket.timeout:
                    return


def get_envelope_token(identifier):
    token = conn.get(
        '{}:token:{}'.format(settings.MAILSEND['CACHE_PREFIX'], identifier))
    if token:
        return token.decode('utf-8')


def set_envelope_token(identifier):
    token = str(uuid.uuid4())
    conn.set(
        '{}:token:{}'.format(settings.MAILSEND['CACHE_PREFIX'], identifier),
        token, settings.MAILSEND['TOKEN_CACHE_TIMEOUT'])
    return token


def delete_envelope_token(identifier):
    return conn.delete('{}:token:{}'.format(
        settings.MAILSEND['CACHE_PREFIX'], identifier))



from datetime import timedelta

from django_redis import get_redis_connection

from . import CACHE_PREFIX
from . import CACHE_TIMEOUT
from . import WorkerPolicyBase

conn = get_redis_connection('default')


class Policy(WorkerPolicyBase):
    """
        {'min_retry': 60 * 5}
    """
    def apply(self, workers):
        latest_status = self.get_latest(self.identifier)

        if not latest_status:
            self.logger.debug(
                '[{}] No previous status that represent a '
                'final state. Nothing to do...'.format(self.identifier))
            return workers

        if self.reply_message and 'greylist' in self.reply_message.lower():
            self.logger.debug(
                '[{}] Greylisting detected in reply message'.format(
                    self.identifier))
            now = self.now()
            for worker in workers:
                if worker.get('ip') == latest_status.get('source_ip'):
                    not_before = now + timedelta(
                        seconds=self.get_settings(worker).get(
                            'min_retry', 60 * 5))
                    self.logger.debug(
                        '[{}] This mail must be sent with {} not '
                        'before {}'.format(
                            self.identifier, worker, not_before.astimezone()))
                    worker['score'] += 0.5 * len(workers)
                    if not_before > worker.get('next_available'):
                        worker['next_available'] = not_before

        return workers

    @staticmethod
    def get_latest(identifier):
        value = conn.get('{}:greylist:{}'.format(
            CACHE_PREFIX, identifier))
        if not value:
            return {}
        splitted_value = value.decode('utf-8').split(':')
        return {
            'source_ip': splitted_value[0], 'creation_date': splitted_value[1]}

    ###########
    # Signals #
    ###########

    @staticmethod
    def mailstatus_pre_save(instance, manager):
        if instance.status in [instance.DELAYED]:
            conn.set('{}:greylist:{}'.format(
                CACHE_PREFIX, instance.mail.identifier),
                '{}:{}'.format(
                    instance.source_ip,
                    instance.creation_date.timestamp()), CACHE_TIMEOUT)



import os

from setuptools import setup
from setuptools import find_packages

base_dir = os.path.dirname(__file__)

about = {}
with open(os.path.join(base_dir, "munch_mailsend", "__about__.py")) as f:
    exec(f.read(), about)

with open(os.path.join(base_dir, "README.rst")) as f:
    long_description = f.read()


def _find_packages(name):
    return [name] + ['{}/{}'.format(name, f) for f in find_packages(name)]


setup(
    name=about["__title__"],
    version=about["__version__"],
    packages=_find_packages('munch_mailsend') + \
        _find_packages('munch_mailsend_tests'),
    long_description=long_description,
    license=about["__version__"],
    description=about["__summary__"],
    author=about["__author__"],
    author_email=about["__email__"],
    url=about["__uri__"],
    install_requires=[],
    extras_require={
        'tests': [
            'flake8==2.5.4',
            'bumpversion==0.5.3',
            'libfaketime==0.2.1']},
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.9',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',  # noqa
    ]
)



import sys

from celery.signals import worker_shutdown
from celery.signals import celeryd_after_setup
from django.conf import settings

from munch.core.utils import get_worker_types
from munch.core.utils import available_worker_types
from munch.core.celery import catch_exception
from munch.core.celery import munch_tasks_router

available_worker_types += ['mx', 'router']
worker_types = get_worker_types()


def add_queues():
    munch_tasks_router.add_queue(
        'router', settings.MAILSEND.get('ROUTING_QUEUE'))
    munch_tasks_router.add_queue(
        'holder', settings.MAILSEND.get('QUEUED_MAIL_QUEUE'))


def register_tasks():
    tasks_map = {
        'router': ['munch_mailsend.tasks.route_envelope'],
        'gc': [
            'munch_mailsend.tasks.ping_workers',
            'munch_mailsend.tasks.check_disabled_workers',
            'munch_mailsend.tasks.dispatch_queued',
            'munch_mailsend.tasks.purge_raw_mail'
        ]
    }
    munch_tasks_router.import_tasks_map(tasks_map, 'munch_mailsend')


@celeryd_after_setup.connect
@catch_exception
def configure_worker(instance, **kwargs):
    from .models import Worker

    if any([t in worker_types for t in ['mx', 'all']]):
        from .tasks import send_email  # noqa
        sys.stdout.write('[mailsend-app] Registering worker as MX...')
        if not settings.MAILSEND.get('SMTP_WORKER_EHLO_AS') or \
                not settings.MAILSEND.get('SMTP_WORKER_SRC_ADDR'):
            raise Exception(
                'Settings must define "SMTP_WORKER_EHLO_AS" '
                'and "SMTP_WORKER_SRC_ADDR"')
        worker, created = Worker.objects.get_or_create(
            ip=settings.MAILSEND.get('SMTP_WORKER_SRC_ADDR'))
        worker.name = kwargs.get('sender')
        worker.policies_settings = settings.MAILSEND.get(
            'WORKER_POLICIES_SETTINGS', {})
        worker.enabled = True
        worker.save()
        queue = settings.MAILSEND.get(
            'MX_WORKER_QUEUE_PREFIX', '').format(ip=worker.ip)
        munch_tasks_router.register_to_queue(queue)
        queue = settings.MAILSEND.get(
            'MX_WORKER_QUEUE_RETRY_PREFIX', '').format(ip=worker.ip)
        munch_tasks_router.register_to_queue(queue)
    if any([t in worker_types for t in ['router', 'all']]):
        from .tasks import route_envelope  # noqa
        sys.stdout.write('[mailsend-app] Registering worker as ROUTER...')
        munch_tasks_router.register_as_worker('router')
    if any([t in worker_types for t in ['gc', 'all']]):
        from .tasks import ping_workers  # noqa
        from .tasks import dispatch_queued  # noqa
        from .tasks import check_disabled_workers  # noqa
        sys.stdout.write(
            '[mailsend-app] Registering worker as GARBAGE COLLECTOR...')
        munch_tasks_router.register_as_worker('gc')


@worker_shutdown.connect
def worker_shutdown(*args, **kwargs):
    from .models import Worker
    sender = kwargs.get('sender')

    if any([t in worker_types for t in ['mx', 'all']]):
        sys.stdout.write('[mailsend-app] Disabling MX worker instance...')
        workers = Worker.objects.filter(
            ip=settings.MAILSEND.get('SMTP_WORKER_SRC_ADDR'), name=sender)
        if not workers:
            raise Exception(
                "Want to disable worker ({} / {}) but can't find it. "
                "Abnormal situation.".format(
                    settings.MAILSEND.get('SMTP_WORKER_SRC_ADDR'), sender))
        if len(workers) > 1:
            raise Exception(
                "Want to disable worker ({} / {}) but found multiple. "
                "Abnormal situation.".format(
                    settings.MAILSEND.get('SMTP_WORKER_SRC_ADDR'), sender))

        workers[0].enabled = False
        workers[0].save()
        sys.stdout.write(
            '[mailsend-app] MX worker disabled and removed from cache. Bye !')



class DKIMException(Exception):
    """Base class for DKIM errors."""
    pass


class InternalError(DKIMException):
    """Internal error in dkim module. Should never happen."""
    pass


class KeyFormatError(DKIMException):
    """Key format error while parsing an RSA public or private key."""
    pass


class MessageFormatError(DKIMException):
    """RFC822 message format error."""
    pass


class ParameterError(DKIMException):
    """Input parameter error."""
    pass


class ValidationError(DKIMException):
    """Validation error."""
    pass


class InvalidCanonicalizationPolicyError(Exception):
    """The c= value could not be parsed."""
    pass


class InvalidTagValueList(Exception):
    pass


class DuplicateTag(InvalidTagValueList):
    pass


class InvalidTagSpec(InvalidTagValueList):
    pass



from django.conf import settings
from django.utils.module_loading import import_string

from munch.core.mail.models import BaseMailStatusManager


def run_policies(mailstatus, method):
    for path in settings.MAILSEND.get('WORKER_POLICIES'):
        try:
            policy = import_string(path)
        except ImportError:
            raise ImportError(
                '{} points to inexistant worker policy'.format(path))
        except TypeError:
            raise TypeError(
                '{} is not a valid WorkerPolicy'.format(path))
        getattr(policy, method)(mailstatus, BaseMailStatusManager)



# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the author be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
#
# Copyright (c) 2008 Greg Hewgill http://hewgill.com
#
# This has been modified from the original software.
# Copyright (c) 2011 William Grant <me@williamgrant.id.au>

import re

from .exceptions import InvalidCanonicalizationPolicyError


__all__ = [
    'CanonicalizationPolicy',
    'InvalidCanonicalizationPolicyError',
    ]


def strip_trailing_whitespace(content):
    return re.sub(br"[\t ]+\r\n", b"\r\n", content)


def compress_whitespace(content):
    return re.sub(br"[\t ]+", b" ", content)


def strip_trailing_lines(content):
    return re.sub(br"(\r\n)*$", b"\r\n", content)


def unfold_header_value(content):
    return re.sub(br"\r\n", b"", content)


class Simple:
    """Class that represents the "simple" canonicalization algorithm."""

    name = "simple"

    @staticmethod
    def canonicalize_headers(headers):
        # No changes to headers.
        return headers

    @staticmethod
    def canonicalize_body(body):
        # Ignore all empty lines at the end of the message body.
        return strip_trailing_lines(body)


class Relaxed:
    """Class that represents the "relaxed" canonicalization algorithm."""

    name = "relaxed"

    @staticmethod
    def canonicalize_headers(headers):
        # Convert all header field names to lowercase.
        # Unfold all header lines.
        # Compress WSP to single space.
        # Remove all WSP at the start or end of the field value (strip).
        canonicalized_headers = []
        for header, value in headers:
            canonicalized_headers.append((
                header.lower().rstrip(),
                compress_whitespace(
                    unfold_header_value(value)).strip() + b'\r\n'))
        return canonicalized_headers

    @staticmethod
    def canonicalize_body(body):
        # Remove all trailing WSP at end of lines.
        # Compress non-line-ending WSP to single space.
        # Ignore all empty lines at the end of the message body.
        return strip_trailing_lines(
            compress_whitespace(strip_trailing_whitespace(body)))


class CanonicalizationPolicy:
    def __init__(self, header_algorithm, body_algorithm):
        self.header_algorithm = header_algorithm
        self.body_algorithm = body_algorithm

    @classmethod
    def from_c_value(cls, canonicalization):
        """Construct the canonicalization policy described by a c= value.

        May raise an C{InvalidCanonicalizationPolicyError} if the given
        value is invalid

        @param c: c= value from a DKIM-Signature header field
        @return: a C{CanonicalizationPolicy}
        """
        if canonicalization is None:
            canonicalization = 'simple/simple'

        splitted_canonicalization = canonicalization.split('/')
        if len(splitted_canonicalization) not in (1, 2):
            raise InvalidCanonicalizationPolicyError(canonicalization)
        if len(splitted_canonicalization) == 1:
            splitted_canonicalization.append('simple')
        headers_canonicalization = splitted_canonicalization[0]
        body_canonicalization = splitted_canonicalization[1]
        try:
            header_algorithm = ALGORITHMS[headers_canonicalization]
            body_algorithm = ALGORITHMS[body_canonicalization]
        except KeyError as err:
            raise InvalidCanonicalizationPolicyError(err.args[0])
        return cls(header_algorithm, body_algorithm)

    def to_c_value(self):
        return b'/'.join((
            self.header_algorithm.name.encode('ascii'),
            self.body_algorithm.name.encode('ascii')))

    def canonicalize_headers(self, headers):
        return self.header_algorithm.canonicalize_headers(headers)

    def canonicalize_body(self, body):
        return self.body_algorithm.canonicalize_body(body)


ALGORITHMS = dict((c.name, c) for c in (Simple, Relaxed))



from functools import wraps

from django.conf import settings
from slimta.envelope import Envelope

from .backoff import *  # noqa


def message_to_envelope(message):
    generated_message = message.message()
    envelope = Envelope()
    envelope.parse(generated_message.as_bytes())
    envelope.sender = message.from_email
    envelope.recipients.append(message.to[0])
    return envelope


def save_timer(name):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            if settings.STATSD_ENABLED:
                from statsd.defaults.django import statsd
                timer = statsd.timer(name)
                timer.start()
            result = f(*args, **kwargs)
            if settings.STATSD_ENABLED:
                timer.stop()
            return result
        return wrapper
    return decorator
